"""
  Test basic driving functionality
"""
import math
from datetime import timedelta

from osgar.node import Node


class Go(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('desired_speed')
        self.start_pose = None
        self.traveled_dist = 0.0
        self.verbose = False
        self.speed = config['max_speed']
        self.dist = config['dist']
        self.timeout = timedelta(seconds=config['timeout'])

    def send_speed_cmd(self, speed, angular_speed):
        return self.publish('desired_speed', [round(speed*1000), round(math.degrees(angular_speed)*100)])

    def update(self):
        channel = super().update()  # define self.time
        if self.verbose:
            print(self.time, 'Go', channel)
        if channel == 'pose2d':
            x, y, heading = self.pose2d
            pose = (x/1000.0, y/1000.0, math.radians(heading/100.0))
            if self.start_pose is None:
                self.start_pose = pose
            self.traveled_dist = math.hypot(pose[0] - self.start_pose[0], pose[1] - self.start_pose[1])

    def wait(self, dt):  # TODO refactor to some common class
        if self.time is None:
            self.update()
        start_time = self.time
        while self.time - start_time < dt:
            self.update()

    def run(self):
        self.update()  # define self.time
        print(self.time, "Go!")
        start_time = self.time
        if self.dist >= 0:
            self.send_speed_cmd(self.speed, 0.0)
        else:
            self.send_speed_cmd(-self.speed, 0.0)
        while self.traveled_dist < abs(self.dist) and self.time - start_time < self.timeout:
            self.update()
        print(self.time, "STOP")
        self.send_speed_cmd(0.0, 0.0)
        self.wait(timedelta(seconds=1))
        print(self.time, "distance:", self.traveled_dist, "time:", (self.time - start_time).total_seconds())


if __name__ == "__main__":
    import argparse
    import os.path
    from osgar.replay import replay
    from osgar.record import record
    from osgar.lib.config import config_load

    parser = argparse.ArgumentParser(description='go')
    subparsers = parser.add_subparsers(help='sub-command help', dest='command')
    subparsers.required = True
    parser_run = subparsers.add_parser('run', help='run on real HW')
    parser_run.add_argument('config', nargs='+', help='configuration file')
    parser_run.add_argument('--note', help='add description')
    parser_run.add_argument('--speed', '-s', help="speed in m/s (default: from config)", type=float)
    parser_run.add_argument('--dist', '-d', help="distance in m (default: %(default)sm)", type=float, default=1.0)
    parser_run.add_argument('--timeout', help='seconds before stopping (default: %(default)s)', type=int, default=10)

    parser_replay = subparsers.add_parser('replay', help='replay from logfile')
    parser_replay.add_argument('logfile', help='recorded log file')
    parser_replay.add_argument('--force', '-F', dest='force', action='store_true', help='force replay even for failing output asserts')
    parser_replay.add_argument('--config', nargs='+', help='force alternative configuration file')
    parser_replay.add_argument('--verbose', '-v', help="verbose mode", action='store_true')

    args = parser.parse_args()
    print(args)

    if args.command == 'replay':
        args.module = 'app'
        game = replay(args, application=Go)
        game.verbose = args.verbose
        game.run()

    elif args.command == 'run':
        prefix = 'go-' + os.path.basename(args.config[0]).split('.')[0] + '-'
        cfg = config_load(*args.config, application=Go)

        # apply overrides from command line
        cfg['robot']['modules']['app']['init']['dist'] = args.dist
        cfg['robot']['modules']['app']['init']['timeout'] = args.timeout
        if args.speed is not None:
            cfg['robot']['modules']['app']['init']['max_speed'] = args.speed

        record(cfg, prefix)

# vim: expandtab sw=4 ts=4



"""
  Spider3 Rider Driver
"""


import struct
from threading import Thread

from osgar.bus import BusShutdownException


CAN_BRIDGE_READY = b'\xfe\x10'  # CAN bridge is ready to accept configuration commands
CAN_BRIDGE_SYNC = b'\xFF'*10    # CAN bridge synchronization bytes
CAN_SPEED_1MB = b'\xfe\x57'     # configure CAN bridge to communicate on 1Mb CAN network
CAN_BRIDGE_START = b'\xfe\x31'  # start bridge


def CAN_packet(msg_id, data):
    header = [(msg_id>>3) & 0xff, (msg_id<<5) & 0xe0 | (len(data) & 0xf)]
    return bytes(header + data)


class Spider(Thread):
    def __init__(self, config, bus):
        bus.register('can', 'status')
        Thread.__init__(self)
        self.setDaemon(True)

        self.bus = bus
        self.buf = b''

        self.can_bridge_initialized = False
        self.status_word = None  # not defined yet
        self.wheel_angles = None  # four wheel angles as received via CAN
        self.zero_steering = None  # zero position of all 4 wheels
        self.speed_cmd = [0, 0]
        self.status_cmd = 3
        self.alive = 0  # toggle with 128
        self.desired_angle = None  # in Spider mode desired weels direction
        self.desired_speed = None

    @staticmethod
    def split_buffer(data):
        # skip 0xFF prefix bytes (CAN bridge control bytes)
        data = data.lstrip(b'\xff')

        if len(data) >= 2:
            # see https://en.wikipedia.org/wiki/CAN_bus
            header = data[:2]
            rtr = (header[1] >> 4) & 0x1  # Remote transmission request
            size = (header[1]) & 0x0f
            if rtr:
                return data[2:], header
            elif len(data) >= 2 + size:
                return data[2+size:], data[:2+size]
        return data, b''  # no complete packet available yet

    @staticmethod
    def fix_range(value):
        "into <-256, +256) interval"
        if value < -256:
            value += 512
        elif value >= 256:
            value -= 512
        return value

    def process_packet(self, packet, verbose=False):
        if packet == CAN_BRIDGE_READY:
            self.bus.publish('can', CAN_BRIDGE_SYNC)
            self.bus.publish('can', CAN_SPEED_1MB)
            self.bus.publish('can', CAN_BRIDGE_START)
            self.can_bridge_initialized = True
            return None

        if len(packet) >= 2:
            msg_id = ((packet[0]) << 3) | (((packet[1]) >> 5) & 0x1f)
            if verbose:
                print(hex(msg_id), packet[2:])
            if msg_id == 0x200:
                self.status_word = struct.unpack('H', packet[2:])[0]
                if self.wheel_angles is not None and self.zero_steering is not None:
                    ret = [self.status_word, [Spider.fix_range(a - b) for a, b in zip(self.wheel_angles, self.zero_steering)]]
                else:
                    ret = [self.status_word, None]

                # handle steering
                if self.desired_angle is not None and self.desired_speed is not None:
                    self.send((self.desired_speed, self.desired_angle))
                else:
                    self.send((0, 0))
                return ret

            elif msg_id == 0x201:
                assert len(packet) == 2 + 8, packet
                self.wheel_angles = struct.unpack_from('HHHH', packet, 2)
                if verbose and self.wheel_angles is not None and self.zero_steering is not None:
                    print('Wheels:',
                          [Spider.fix_range(a - b) for a, b in zip(self.wheel_angles, self.zero_steering)])
            elif msg_id == 0x203:
                assert len(packet) == 2 + 8, packet
                prev = self.zero_steering
                self.zero_steering = struct.unpack_from('HHHH', packet, 2)
                if verbose:
                    print('Zero', self.zero_steering)
                # make sure that calibration did not change during program run
                assert prev is None or prev == self.zero_steering, (prev, self.zero_steering)
            elif msg_id == 0x204:
                assert len(packet) == 2 + 8, packet
                val = struct.unpack_from('HHBBH', packet, 2)
                if verbose:
                    print("User:", val[2]&0x7F, val[3]&0x7F, val)

    def process_gen(self, data, verbose=False):
        self.buf, packet = self.split_buffer(self.buf + data)
        while len(packet) > 0:
            ret = self.process_packet(packet, verbose=verbose)
            if ret is not None:
                yield ret
            self.buf, packet = self.split_buffer(self.buf)  # i.e. process only existing buffer now

    def run(self):
        try:
            while True:
                dt, channel, data = self.bus.listen()
                if channel == 'raw':
                    if len(data) > 0:
                        for status in self.process_gen(data):
                            if status is not None:
                                self.bus.publish('status', status)
                elif channel == 'move':
                    self.desired_speed, self.desired_angle = data
                else:
                    assert False, channel  # unsupported channel
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()

    def send(self, data):
        if self.can_bridge_initialized:
            speed, angular_speed = data
            if speed > 0:
                if self.status_word is None or self.status_word & 0x10 != 0:
                    angle_cmd = int(angular_speed)  # TODO verify angle, byte resolution
                else:
                    print('SPIDER MODE')
                    desired_angle = int(angular_speed)  # TODO proper naming etc.
                    if self.wheel_angles is not None and self.zero_steering is not None:
                        curr = Spider.fix_range(self.wheel_angles[0] - self.zero_steering[0])
                        diff = Spider.fix_range(desired_angle - curr)
                        print('DIFF', diff)
                        if abs(diff) < 5:
                            angle_cmd = 0
                        elif diff < 0:
                            angle_cmd = 50
                        else:
                            angle_cmd = 0x80 + 50
                    else:
                        angle_cmd = 0
                if speed >= 10:
                    packet = CAN_packet(0x401, [0x80 + 127, angle_cmd])
                else:
                    packet = CAN_packet(0x401, [0x80 + 80, angle_cmd])
            else:
                packet = CAN_packet(0x401, [0, 0])  # STOP packet
            self.bus.publish('can', packet)

            # alive
            packet = CAN_packet(0x400, [self.status_cmd, self.alive])
            self.bus.publish('can', packet)
            self.alive = 128 - self.alive
        else:
            print('CAN bridge not initialized yet!')
#            self.logger.write(0, 'ERROR: CAN bridge not initialized yet! [%s]' % str(data))

# vim: expandtab sw=4 ts=4



#!/usr/bin/python
"""
  Extract OSGAR logfile from SubT ROS Bag (topic /robot_data)
"""
# based on format specification at
#   http://wiki.ros.org/Bags/Format/2.0
import os
import struct
import tarfile

MAX_RECORD_SIZE = 100000000

OP_MESSAGE_DATA = 0x2
OP_INDEX_DATA = 0x4
OP_CHUNK = 0x5
OP_CHUNK_INFO = 0x6
OP_CONNECTION = 0x7

IMAGE_TOPIC_ID = 59  # /usb_cam/image_raw


def parse_header(header):
    index = 0
    ret = {}
    while index < len(header):
        size = struct.unpack_from('<I', header, index)[0]
        assert size > 0, size
        record = header[index + 4 : index + 4 + size]
        assert b'=' in record, record
        key = record.split(b'=')[0].decode('ascii')
        value = record[len(key)+1:]  # potentially binary blob
        ret[key] = value  # by default keep original binary data
        if key == 'op':
            assert len(value) == 1, len(value)
            # bag header, chunk, index, connection, chunk info
            assert value[0] in [0x2, 0x3, 0x5, 0x4, 0x7, 0x6], value[0]
            ret[key] = value[0]
        elif key == 'conn':
            assert len(value) == 4, len(value)
            ret[key] = struct.unpack('<I', value)[0]
        elif key == 'count':
            assert len(value) == 4, len(value)
            ret[key] = struct.unpack('<I', value)[0]
        elif key == 'topic':
            # actually expected only in OP=0x7 (connection)
            pass
        elif key == 'compression':
            assert value == b'none', value  # note, that b'bz2' is also supported
        elif key == 'time':
            assert len(value) == 8, len(value)
            ret[key] = struct.unpack('<II', value)
        index += 4 + size
    assert index == len(header), (index, len(header))
    return ret


def handle_message_data(index_header_dict, data):
    # verify index header and split block data into new header and data
    header_size = struct.unpack_from('<I', data)[0]
    header_dict = parse_header(data[4:4+header_size])
    size = struct.unpack_from('<I', data, 4+header_size)[0]
    return header_dict, data[4+header_size:4+header_size+4+size]


def read_rosbag_fd_gen(f, verbose=False):
    """Read ROS Bag with already opened file descriptor f"""
    header = f.read(13)
    assert header == b'#ROSBAG V2.0\n', header

    header_len = f.read(4)
    chunk = None
    while len(header_len) == 4:
        size = struct.unpack('<I', header_len)[0]
        assert size < MAX_RECORD_SIZE, size
        header = f.read(size)
        assert len(header) == size, (len(header), size)
        header_dict = parse_header(header)
        if 'topic' in header_dict and verbose:
            print('topic', header_dict['conn'], header_dict['topic'])

        data_len = f.read(4)
        assert len(data_len) == 4, len(data_len)
        size = struct.unpack('<I', data_len)[0]
        assert size < MAX_RECORD_SIZE, size
        data = f.read(size)
        #assert len(data) == size, (len(data), size)
        if len(data) != size:
            print("Error", (len(data), size))
            return
        op = header_dict['op']
        if op == OP_CHUNK:
            chunk = data
        elif op == OP_INDEX_DATA:
            header_count = header_dict['count']
            assert chunk is not None  # first data then indexes
            # note, that one "chunk" can be shared by several "indexes"
            assert size == 12 * header_count, size
            for t, offset in struct.iter_unpack('<QI', data):
                assert offset < len(chunk)
                yield handle_message_data(header_dict, chunk[offset:])
        else:
            yield header_dict, data_len + data
        header_len = f.read(4)


def read_rosbag_gen(filename):
    with open(filename, 'rb') as f:
        for data in read_rosbag_fd_gen(f):
            yield  data


def parse_raw_image(data, dump_filename=None):
    # http://docs.ros.org/melodic/api/sensor_msgs/html/msg/Image.html
    size = struct.unpack_from('<I', data)[0]
    assert size == 2764848, size  # expected size for raw image during experiment
    # http://docs.ros.org/melodic/api/std_msgs/html/msg/Header.html
    pos = 4
    seq, timestamp_sec, timestamp_nsec, frame_id_size = struct.unpack_from('<IIII', data, pos)
    pos += 4 + 4 + 4 + 4
    frame_id = data[pos:pos+frame_id_size]
    print(frame_id, timestamp_sec, timestamp_nsec)
    pos += frame_id_size
    height, width, encoding_size = struct.unpack_from('<III', data, pos)
    pos += 4 + 4 + 4
    encoding = data[pos:pos+encoding_size]
    pos += encoding_size
    print(height, width, encoding)
    is_bigendian, step, image_arr_size = struct.unpack_from('<BII', data, pos)
    pos += 1 + 4 + 4
    if dump_filename is not None:
        with open(dump_filename, 'wb') as f:
            f.write(b'P6\n%d %d\n255\n' % (width, height))
            f.write(data[pos:pos+image_arr_size])


def parse_string(data, dump_filename=None):
    # http://docs.ros.org/melodic/api/std_msgs/html/msg/Header.html
    size = struct.unpack_from('<I', data)[0]
#    print(size)
    pos = 4
    str_size = struct.unpack_from('<I', data, pos)[0]
#    print(size, str_size)
#    assert size == str_size + 4, (size, str_size)
    pos += 4
    return data[pos:]


def extract_log(gen, out_name, append=False, verbose=False):
    with open(out_name, 'ab' if append else 'wb') as f:
        for header_dict, data in gen:
            op = header_dict['op']
            conn  = header_dict.get('conn')
            if conn == 0:
                if op == OP_MESSAGE_DATA:
                    d = parse_string(data)
                    if b'Hello' not in d:
                        f.write(d)
                    elif verbose:
                        print(d, len(d))


def process_tar(filepath, all, verbose):
        # In: "ver41p3-91c332d3-2066-466c-a9b9-e3418bbeb0a9-A10F900L.tar"
        # Out: "aws-ver41p3-A10F900L.log"
        basename_without_ext = os.path.basename(filepath).split('.', 1)[0]
        s = basename_without_ext.split('-')
        robot_name = s[-2] if s[-1] == "commsbridge" else s[-1]
        name = robot_name + ".log"
        letter = robot_name[0]
        out_name = os.path.join(os.path.dirname(filepath), name)
        robot_data_0_processed = False
        robot_data_1_processed = False
        with tarfile.open(filepath, "r") as tar:
            for member in tar.getmembers():
                if member.name.startswith('robot_data_0.bag'):
                    print(member.name, "->", name)
                    f = tar.extractfile(member)
                    extract_log(read_rosbag_fd_gen(f, verbose=verbose), out_name, verbose=verbose)
                    robot_data_0_processed = True
                elif member.name.startswith('robot_data_1.bag'):
                    if not all:
                        print(member.name, "->", "SKIPPED")
                        continue
                    print(member.name, "->", name)
                    assert robot_data_0_processed
                    f = tar.extractfile(member)
                    extract_log(read_rosbag_fd_gen(f, verbose=verbose), out_name, append=True, verbose=verbose)
                    robot_data_1_processed = True
                elif member.name.startswith('robot_data_2.bag'):
                    if not all:
                        print(member.name, "->", "SKIPPED")
                        continue
                    print(member.name, "->", name)
                    assert robot_data_0_processed
                    assert robot_data_1_processed
                    f = tar.extractfile(member)
                    extract_log(read_rosbag_fd_gen(f, verbose=verbose), out_name, append=True, verbose=verbose)
                elif member.name.endswith('rosout.log'):
                    rosout_name = letter + '-rosout.log'
                    print(member.name, "->", rosout_name)
                    with open(os.path.join(os.path.dirname(filepath), rosout_name), 'wb') as f:
                        f.write(tar.extractfile(member).read())
                elif (member.name in ['server_console.log'] or member.name.endswith('.yml') or
                      member.name.startswith('subt_urban_') or member.name.startswith('subt_cave_') or
                      (member.name.startswith('state.tlog') and all)):
                    print(member.name)
                    with open(os.path.join(os.path.dirname(filepath), member.name.replace(':', '_')), 'wb') as f:
                        f.write(tar.extractfile(member).read())


def main():
    import argparse
    import pathlib
    import sys

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('filenames', help='ROS bag file(s) or tar(.gz)', nargs='*')
    #parser.add_argument('-a', '--all', help='extract all robot_data* files', action='store_true')
    parser.add_argument('--debug', help='debug output', action='store_true')
    args = parser.parse_args()

    if args.filenames == []:
        for p in pathlib.Path('.').iterdir():
            if p.suffixes == ['.tar', '.gz'] or p.suffixes == ['.tar']:
                args.filenames.append(str(p))
        if len(args.filenames) == 0:
            sys.exit("no logfiles found in current directory")
        args.filenames.sort()

    for filename in args.filenames:
        print("processing:", filename)
        if filename.endswith('.tar') or filename.endswith('.tar.gz'):
            process_tar(filename, all=True, verbose=args.debug)
        else:
            out_name = os.path.join(os.path.dirname(filename), 'tmp.log')
            extract_log(read_rosbag_gen(filename), out_name, verbose=args.debug)


if __name__ == "__main__":
    main()
# vim: expandtab sw=4 ts=4 




import unittest
from unittest.mock import patch
import math

from subt.report_artf import triple, score


class ReportArtfTest(unittest.TestCase):

    def test_triple(self):
        arr = triple(0, 0, 0)
        self.assertEqual(len(arr), 3)
        self.assertEqual(arr[0], (4, 0, 0))
        self.assertAlmostEqual(arr[1][0], -2)
        self.assertAlmostEqual(arr[1][1], math.sqrt(12))
        self.assertAlmostEqual(arr[2][0], -2)
        self.assertAlmostEqual(arr[2][1], -math.sqrt(12))

        arr = triple(10, 0, 0)
        self.assertEqual(len(arr), 3)
        self.assertEqual(arr[0], (14, 0, 0))


    def test_score(self):
        with patch('subt.report_artf.get_status') as p, \
             patch('subt.report_artf.report_artf') as p2, \
             patch('time.sleep'):
            p.side_effect = [
                b'{"score":0,"remaining_reports":38,"current_team":"robotika","run_clock":983.1}',
                b'{"score":1,"remaining_reports":37,"current_team":"robotika","run_clock":987.4}',
            ]
            self.assertTrue(score('Survivor', 0, 0, 0))

            p.side_effect = [
                b'{"score":0,"remaining_reports":39,"current_team":"robotika","run_clock":966.5}',
                b'{"score":0,"remaining_reports":38,"current_team":"robotika","run_clock":970.8}',
            ]
            self.assertFalse(score('Backpack', 1, 2, 3))

# vim: expandtab sw=4 ts=4




"""
   Report artifact for STIX in Colorado April 2019
         reused for Pittsburgh August 2019
     and reused again in Elma/Olympia February 2020
"""
import requests
import time
import math
import json


URL_BASE = "http://localhost:8888"  # local Robotika test/demo
#URL_BASE = "http://10.100.1.200:8000"  # Alpha (was Army and Safety Research) Tunnel
#URL_BASE = "http://10.100.2.200:8000"  # Beta (was Miami and Experimental) Tunnel

ARTF_TYPES = ['Survivor', 'Backpack', 'Cell Phone',  # common
              'Drill', 'Fire Extinguisher',          # tunnel extra
              'Gas', 'Vent']                         # urban extra
ARTF_TYPES_SHORT = [x[0] for x in ARTF_TYPES]

json_headers = {
#    "Authorization" : "Bearer subttesttoken123",  # demo
#    "Authorization" : "Bearer NfEwAHEYsKqQkxSf",  # STIX
    "Content-Type" : "application/json",
}


def get_status():
    print('Get Status')
    url = URL_BASE + "/api/status/"

    # Correct GET /api/status/ request
    response = requests.get(url, headers=json_headers)
    assert response.status_code == 200, response.status_code
    print(response.content)
    print("-------------------")
    return response.content


def report_artf(artf_type, x, y, z):
    artifact_report_data = {
        "x": x,
        "y": y,
        "z": z,
        "type": artf_type,
    }
    print('Report', artifact_report_data)
    url = URL_BASE + "/api/artifact_reports/"

    # Correct POST /api/artifact_reports/ request
    response = requests.post(url, json=artifact_report_data, headers=json_headers)
    print(response.content)
    assert response.status_code == 201, response.status_code
    print("-------------------")
    return response.content


def triple(x, y, z):
    """
    Generate 3 coordinates with the same Z and XY in triangle of 4m
    """
    dist = 4.0
    arr = []
    for angle_deg in [0, 120, 240]:
        angle = math.radians(angle_deg)
        arr.append((x + math.cos(angle)*dist, y + math.sin(angle)*dist, z))
    return arr


def score(artf_type, x, y, z):
    """
    return True for scored artifact report
    """
    before = json.loads(bytes.decode(get_status()))
    time.sleep(2)
    report_artf(artf_type, x, y, z)
    time.sleep(2)
    after = json.loads(bytes.decode(get_status()))
    return before['score'] < after['score']


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='Report artifact to server')
    parser.add_argument('artf_type', help='Type of artifact', choices=ARTF_TYPES + ARTF_TYPES_SHORT)
    parser.add_argument('x', help='X coordinate in meters', type=float)
    parser.add_argument('y', help='Y coordinate in meters', type=float)
    parser.add_argument('z', help='Z coordinate in meters', type=float)
#    parser.add_argument('--only-one', '-1', help='only one exact shot',
#                        action='store_true')
    args = parser.parse_args()

    artf_type = args.artf_type
    if artf_type in ARTF_TYPES_SHORT:
        artf_type = ARTF_TYPES[ARTF_TYPES_SHORT.index(artf_type)]

    print('Reporting:', artf_type)
    if True:  # args.only_one:
        print(score(artf_type, args.x, args.y, args.z))
    else:
        for x, y, z in triple(args.x, args.y, args.z):
            if score(artf_type, x, y, z):
                break

# vim: expandtab sw=4 ts=4




"""
  Convert depth image to "lidar" scan
"""
import logging
from datetime import timedelta
import numpy as np

from osgar.node import Node

g_logger = logging.getLogger(__name__)


class ScanMixer(Node):

    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register("scan")
        self.scan = None  # original lidar scan
        self.rs_scan = None  # RealSense RGBD converted scan from ROS
        self.rs_scan_time = timedelta(0)
        self.verbose = False

    def update(self):
        channel = super().update()
        assert channel in ["scan", "rs_scan"], channel

        if channel == 'rs_scan':
            self.rs_scan_time = self.time
        elif channel == 'scan':
            #assert len(self.scan) in [271, ], len(self.scan)  # Eduro=271
            if self.rs_scan is not None:
                max_delay = timedelta(milliseconds=500)
                rs_delay = self.time - self.rs_scan_time
                if rs_delay > max_delay:
                    g_logger.debug(f"rs_scan older than {max_delay}: {rs_delay}")
                    self.publish('scan', self.scan)
                else:
                    indexes = np.linspace(0, len(self.rs_scan), num=len(self.scan), endpoint=False, dtype=int)
                    rs_scan = np.asarray(self.rs_scan)
                    scan = np.asarray(self.scan)
                    # replace infinity with long distance
                    infinity = 30000
                    rs_scan[rs_scan == 0] = infinity
                    scan[scan == 0] = infinity
                    # resample
                    rs_scan = rs_scan[indexes]
                    scan = np.minimum(rs_scan, scan)
                    # replace long distance with infinity
                    scan[scan == infinity] = 0
                    self.publish('scan', scan.tolist())
            else:
                self.publish('scan', self.scan)
        else:
            assert False, channel  # unsupported channel

        return channel


if __name__ == "__main__":
    import argparse
    from unittest.mock import MagicMock
    from osgar import logger
    from osgar import bus
    from osgar.lib.serialize import deserialize

    parser = argparse.ArgumentParser(description='')
    parser.add_argument('logfile', help='filename of stored file')
    rs_scan = "rosmsg.scan"
    scan = "lidar.scan"
    stream_names = [rs_scan, scan]

    bus = bus.Bus(MagicMock())
    tester = bus.handle('tester')
    tester.register("scan", "rs_scan")
    mixer = ScanMixer(config={}, bus=bus.handle('mixer'))
    bus.connect("tester.scan", "mixer.scan")
    bus.connect("tester.rs_scan", "mixer.rs_scan")
    bus.connect("mixer.scan", "tester.mixed")
    mixer.start()

    args = parser.parse_args()
    stream_ids = [logger.lookup_stream_id(args.logfile, stream) for stream in stream_names]
    id2name = { stream_id: name for stream_id, name in zip(stream_ids, ["rs_scan", "scan"]) }
    with logger.LogReader(args.logfile, only_stream_id=stream_ids) as log:
        for timestamp, stream_id, bytes_data in log:
            data = deserialize(bytes_data)
            tester.publish(id2name[stream_id], data)
            if id2name[stream_id] == "scan":
                t, s, d = tester.listen()

    mixer.request_stop()
    mixer.join()

# vim: expandtab sw=4 ts=4



import contextlib

from threading import Thread

import zmq
import msgpack

from osgar.bus import BusShutdownException


class Push:

    def __init__(self, config, bus):
        bus.register()
        self.endpoint = config.get('endpoint', 'tcp://127.0.0.1:5566')
        self.timeout = config.get('timeout', 1) # default send timeout 1s, effective on full queue only
        self.thread = Thread(target=self.run)
        self.thread.name = bus.name
        self.bus = bus

    def start(self):
        self.thread.start()

    def join(self, timeout=None):
        self.thread.join(timeout=timeout)

    def run(self):
        context = zmq.Context.instance()
        socket = context.socket(zmq.PUSH)

        # https://stackoverflow.com/questions/7538988/zeromq-how-to-prevent-infinite-wait
        socket.SNDTIMEO = int(self.timeout * 1000)  # convert to milliseconds
        socket.LINGER = 100 #milliseconds
        socket.connect(self.endpoint)

        try:
            with contextlib.closing(socket):
                while True:
                    dt, channel, data = self.bus.listen()
                    print("*"*50)
                    print("sending:", data)
                    raw = msgpack.packb(data, use_bin_type=True)
                    socket.send_multipart([bytes(channel, 'ascii'), raw])
        except zmq.ZMQError as e:
            if e.errno == zmq.EAGAIN:
                pass #TODO log timeout
            else:
                pass #TODO log other error
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()



import unittest
from unittest.mock import patch, MagicMock, call

from osgar.drivers.logsocket import LogTCPStaticIP as LogTCP, LogTCPDynamicIP, LogTCPServer, LogUDP, LogHTTP
from osgar.bus import Bus
import osgar.node

class LogSocketTest(unittest.TestCase):

    def test_tcp_send(self):
        with patch('osgar.drivers.logsocket.socket.socket') as mock:
            instance = mock.return_value

            logger = MagicMock()
            logger.register = MagicMock(return_value=1)
            bus = Bus(logger)
            config = {'host': '192.168.2.23', 'port':2111}
            device = LogTCP(config=config, bus=bus.handle('tcp'))
            tester = bus.handle('tester')
            tester.register('raw')
            bus.connect('tester.raw', 'tcp.raw')
            tester.publish('raw', b'bin data')
            device.start()
            device.request_stop()
            device.join()
            instance.send.assert_called_once_with(b'bin data')


    def test_udp_send(self):
        with patch('osgar.drivers.logsocket.socket.socket') as mock:
            instance = mock.return_value

            logger = MagicMock()
            bus = Bus(logger)
            config = {'host': '192.168.2.23', 'port':2111}
            device = LogUDP(config=config, bus=bus.handle('udp'))
            tester = bus.handle('tester')
            tester.register('raw')
            bus.connect('tester.raw', 'udp.raw')
            tester.publish('raw', b'bin data')
            device.start()
            device.request_stop()
            device.join()
            instance.sendto.assert_called_once_with(
                    b'bin data', ('192.168.2.23', 2111))

    def test_tcp_server(self):
        with patch('osgar.drivers.logsocket.socket.socket') as mock:
            instance = mock.return_value
            accept = MagicMock()
            accept.recv = MagicMock(return_value=b'some bin data')
            instance.accept = MagicMock(return_value=(accept, ('127.0.0.1', 1234)))

            logger = MagicMock()
            bus = Bus(logger)
            config = {'host': '192.168.1.2', 'port':8080}
            device = LogTCPServer(config=config, bus=bus.handle('tcp'))
            tester = bus.handle('tester')
            bus.connect('tcp.raw', 'tester.raw')

            device.start()
            tester.listen()
            device.request_stop()
            device.join()

            instance.listen.assert_called_once_with(1)
            instance.bind.assert_called_once_with(('192.168.1.2', 8080))
            instance.accept.assert_called_once_with()
            accept.recv.assert_called_with(device.bufsize)

    def test_dynamic_tcp(self):
        with patch('osgar.drivers.logsocket.socket.socket') as mock:
            instance = mock.return_value

            logger = MagicMock()
            bus = Bus(logger)
            device = LogTCPDynamicIP(config={}, bus=bus.handle('tcpdyn'))
            tester = bus.handle('tester')
            tester.register('addr')
            bus.connect('tester.addr', 'tcpdyn.addr')
            tester.publish('addr', ['10.1.10.1', 8000])
            device.start()
            device.request_stop()
            device.join()

            instance.connect.assert_called_once_with(('10.1.10.1', 8000))

    def test_not_started_dynamic_tcp(self):
        with patch('osgar.drivers.logsocket.socket.socket') as mock:
            instance = mock.return_value

            logger = MagicMock()
            bus = Bus(logger)
            config = {}
            device = LogTCPDynamicIP(config=config, bus=bus.handle('tcp'))
            device.start()
            device.request_stop()
            device.join()

            instance.connect.assert_not_called()

    def test_dynamic_tcp_called_twice(self):
        with patch('osgar.drivers.logsocket.socket.socket') as mock:
            instance = mock.return_value

            logger = MagicMock()
            bus = Bus(logger)
            device = LogTCPDynamicIP(config={}, bus=bus.handle('tcpdyn'))
            tester = bus.handle('tester')
            tester.register('addr')
            bus.connect('tester.addr', 'tcpdyn.addr')
            tester.publish('addr', ['10.1.10.1', 8000])
            tester.publish('addr', ['192.168.1.31', 8010])
            device.start()
            tester.sleep(0.1)
            device.request_stop()
            device.join()

            self.assertEqual(instance.connect.call_args_list, [
                    call(('10.1.10.1', 8000)),
                    call(('192.168.1.31', 8010))
                ])

    def test_http_sleep(self):
        # reported as bug for IP camera running at full speed
        with patch('osgar.drivers.logsocket.urllib.request.urlopen') as mock:
            instance = mock.return_value
            instance.__enter__.return_value.read = MagicMock(return_value=b'123')
            logger = MagicMock()
            logger.register = MagicMock(return_value=1)
            logger.write = MagicMock(return_value=123)
            bus = Bus(logger)
            config = {
              "url": "http://192.168.0.99/img.jpg",
              "sleep": 0.1,
              "timeout": 1.0
            }
            device = LogHTTP(config=config, bus=bus.handle('http'))
            tester = bus.handle('tester')
            bus.connect('http.raw', 'tester.raw')

            device.start()
            data = tester.listen()
            self.assertEqual(data, (123, 'raw', b'123'))
            device.request_stop()
            device.join()
            self.assertEqual(
                    len(instance.__enter__.return_value.read.call_args_list),
                    1)  # it should be just one call and sleep


class Echo(osgar.node.Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('raw0', 'raw1')
        self.data = config['data']

    def run(self):
        self.bus.sleep(0.1) # give other nodes a chance to enter listen call
        self.bus.publish('raw0', self.data)
        dt, channel, data = self.bus.listen()
        assert data == b'ok'
        self.bus.publish('raw1', self.data)
        dt, channel, data = self.bus.listen()
        assert data == b'ok'


if __name__ == "__main__":
    import os
    os.environ['OSGAR_LOGS'] = '.'
    from osgar.zmqrouter import record
    config = {'version': 2, 'robot': {'modules': {}, 'links': []}}
    config['robot']['modules'] = {
        "0": {
            "driver": "udp",
            "init": {"host": "127.0.0.1", "port": 9111, "timeout": 0.5},
        },
        "1": {
            "driver": "udp",
            "init": {"host": "127.0.0.1", "port": 9111, "timeout": 0.5},
        },
        "echo": {
            "driver": "osgar.drivers.test_logsocket:Echo",
            "init": {"data": b"ok" }
        },
    }
    config['robot']['links'] = [
        ['echo.raw0', '0.raw'],
        ['1.raw', 'echo.stdout'],
        ['echo.raw1', '1.raw'],
        ['0.raw', 'echo.stdout']
    ]
    record(config, log_filename='logzeromq-zmqrouter.log')


# vim: expandtab sw=4 ts=4



"""
  Virtual Bumper
  - origin from 2002 Cleaning Contest
  - if robot is not moving and receive commands to move then it is stuck probably
"""
import math

from osgar.lib.mathex import normalizeAnglePIPI


def equal_poses(pose1, pose2, dist_limit, angle_limit=None):
    x1, y1, heading1 = pose1
    x2, y2, heading2 = pose2
    return (math.hypot(x2 - x1, y2 - y1) < dist_limit and 
            (angle_limit is None or abs(normalizeAnglePIPI(heading1 - heading2)) < angle_limit))


class VirtualBumper:
    def __init__(self, stuck_time_limit, dist_radius_limit, angle_limit=None):
        self.stuck_time_limit = stuck_time_limit
        self.dist_radius_limit = dist_radius_limit
        self.angle_limit = angle_limit
        self.should_be_moving = False  # status based on desired speed commands
        self.verbose = False
        self.reset_counters()

    def reset_counters(self):
        self.last_pose = None
        self.last_pose_time = None  # time of last pose without motion
        self.not_moving = None  # how long it is not moving?

    def update_desired_speed(self, desired_speed, desired_angular_speed):
        self.should_be_moving = abs(desired_speed) > 0.001 or abs(desired_angular_speed) > 0.001
        if self.verbose:
            print('VirtualBumper', desired_speed, desired_angular_speed, self.should_be_moving)
        if not self.should_be_moving:
            self.reset_counters()

    def update_pose(self, timestamp, pose):
        if self.verbose:
            print('VirtualBumper', timestamp, pose)
        if self.last_pose is not None:
            if equal_poses(pose, self.last_pose, self.dist_radius_limit, self.angle_limit):
                self.not_moving = timestamp - self.last_pose_time
            else:
                self.last_pose = pose
                self.last_pose_time = timestamp
        else:
            self.last_pose = pose
            self.last_pose_time = timestamp
        if self.verbose:
            print(self.last_pose, self.last_pose_time, self.not_moving, self.stuck_time)
        # if it should not be moving reset all counters
        if not self.should_be_moving:
            self.reset_counters()

    def collision(self):
        if self.not_moving is None:
            return False
        return self.not_moving >= self.stuck_time_limit

# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import MagicMock
import time

import zmq

from osgar.drivers.logzeromq import LogZeroMQ
from osgar.bus import BusShutdownException


class LogZeroMQTest(unittest.TestCase):
    def test_recv_timeout(self):
        config = {
            'mode': 'PULL',
            'endpoint': 'tcp://localhost:5555',
            'timeout': 0.1
        }
        bus = MagicMock()
        bus.is_alive = MagicMock(return_value=True)
        bus.name = "test_recv_timeout"
        node = LogZeroMQ(config, bus)
        node.start()
        time.sleep(0.01)  # give it a chance to start
        node.request_stop()
        bus.is_alive = MagicMock(return_value=False)  # supplement mock request_stop()
        node.join()

    def test_send_timeout(self):
        config = {
            'mode': 'PUSH',
            'endpoint': 'tcp://localhost:5556',
            'timeout': 0.1
        }
        bus = MagicMock()
        bus.listen = MagicMock(return_value=(1, 2, b'data'))
        bus.is_alive = MagicMock(return_value=True)
        node = LogZeroMQ(config, bus)
        node.start()
        time.sleep(0.01)  # give it a chance to start
        node.request_stop()
        bus.listen = MagicMock(side_effect=BusShutdownException())
        bus.is_alive = MagicMock(return_value=False)  # supplement mock request_stop()
        node.join()

    def test_req(self):
        config = {
            'mode': 'REQ',
            'endpoint': 'tcp://localhost:5557',
            'timeout': 0.1
        }
        context = zmq.Context()
        socket = context.socket(zmq.REP)
        socket.RCVTIMEO = 1000
        socket.bind('tcp://127.0.0.1:5557')

        bus = MagicMock()
        bus.listen = MagicMock(return_value=(1, 2, ['abcd', 'set_brakes on\n']))
        bus.is_alive = MagicMock(return_value=True)
        node = LogZeroMQ(config, bus)
        node.start()
        time.sleep(0.2)  # let it start and timeout in recv call
        node.request_stop()
        bus.listen = MagicMock(side_effect=BusShutdownException())
        bus.is_alive = MagicMock(return_value=False)  # supplement mock request_stop()

        msg = socket.recv()
        node.join()
        self.assertEqual(msg, b'set_brakes on\n')
        bus.publish.assert_called_with('timeout', ['abcd', True])

# vim: expandtab sw=4 ts=4



"""
  Driver for robot Robik from cortexpilot.com
"""

import ctypes
import struct
import math
from datetime import timedelta

from osgar.node import Node
from osgar.bus import BusShutdownException
from osgar.lib import quaternion


# CPR = 9958 (ticks per revolution)
# wheel diameter D = 395 mm
# 1 Rev = 1241 mm
ENC_SCALE = 1.241/9958
WHEEL_DISTANCE = 0.88  # meters TODO confirm
RAMP_STEP = 0.1  # fractional number for speed in -1.0 .. 1.0


def sint32_diff(a, b):
    return ctypes.c_int32(a - b).value


class Cortexpilot(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('raw', 'encoders', 'emergency_stop', 'pose2d', 
                     'voltage', 'rotation', 'orientation', 'scan')

        self._buf = b''

        # commands
        self.desired_speed = 0.0  # m/s
        self.desired_angular_speed = 0.0
        self.cmd_flags = 0x40 #0x41  # 0 = remote steering, PWM OFF, laser ON, TODO
        self.speeds = self.plain_speeds()

        # status
        self.emergency_stop = None  # uknown state
        self.pose = (0.0, 0.0, 0.0)  # x, y in meters, heading in radians (not corrected to 2PI)
        self.flags = None
        self.last_encoders = None
        self.yaw = None
        self.lidar_valid = False
        self.lidar_timestamp = 0
        self.uptime = None

    def send_pose(self):
        x, y, heading = self.pose
        self.publish('pose2d', [round(x*1000), round(y*1000),
                                round(math.degrees(heading)*100)])

    def query_version(self):
        ret = bytes([0, 0, 3, 0x1, 0x01])
        checksum = sum(ret) & 0xFF
        return ret + bytes([(256-checksum) & 0xFF])

    def oscilate(self):
        while True:
            end = self.time + timedelta(seconds=1)
            while self.time < end:
                yield self.desired_speed, -self.desired_angular_speed
            end = self.time + timedelta(seconds=1)
            while self.time < end:
                yield -self.desired_speed, -self.desired_angular_speed

    def plain_speeds(self):
        while True:
            yield self.desired_speed, -self.desired_angular_speed

    def create_packet(self):
        if self.yaw is None:
            self.yaw = 0.0  # hack!

        speed_frac, speed_dir = next(self.speeds)
        speed_frac *= 2
        speed_dir *= 1.2

        if speed_frac < 0:
            speed_dir = -speed_dir  # Robik V5.1.1 handles backup backwards
        if not self.lidar_valid:
            speed_frac = 0.0
            speed_dir = 0.0

        #print(self.time, "{:.4f}, {:.4f} \t {:.4f} {:.4f}".format(speed_frac, speed_dir, self.desired_speed, self.desired_angular_speed))

        flags = self.cmd_flags
        flags |= (1<<8)  # agresive turning
        if self.emergency_stop is not None:
            if self.emergency_stop:
                flags |= (1<<11)  # display red LEDs
            else:
                flags |= (1<<10)  # turn on green (9th bit)

        packet = struct.pack('<ffI', speed_frac, speed_dir, flags)
        assert len(packet) < 256, len(packet)  # just to use LSB only
        ret = bytes([0, 0, len(packet) + 2 + 1, 0x1, 0x0D]) + packet
        # addr=0x1, cmd=0xD, length is given by payload, addr, cmd and checksum
        checksum = sum(ret) & 0xFF
        return ret + bytes([(256-checksum) & 0xFF])

    def get_packet(self):
        """extract packet from internal buffer (if available otherwise return None"""
        data = self._buf
        if len(data) < 3:
            return None
        high, mid, low = data[:3]  # 24bit packet length (big endian int)
        assert high == 0, high  # all messages < 65535 bytes
        size = 256 * mid + low + 3  # counting also 3 bytes of packet length header
        if len(data) < size:
            return None
        ret, self._buf = data[:size], data[size:]
        checksum = sum(ret) & 0xFF
        assert checksum == 0, checksum  # checksum error
        return ret

    def parse_packet(self, data):
        """
        Parse cortexpilot sensors status message
        """
        # expects already validated single sample with 3 bytes length prefix
        #   and checksum at the end
        high, mid, low = data[:3]
        assert high == 0, high  # fixed packet size 2*256+89 bytes
        assert mid == 2, mid
        assert low == 89, low
        addr, cmd = data[3:5]
        assert addr == 1, addr
        assert cmd == 0xD, cmd
        offset = 5  # payload offset

        # 4 byte Flags (unsigned long)  0
        #   bit 0 -> 1 = BigRedSwitch
        #   bit 1 -> 1 = MissionButton
        #   bit 2 -> copy of EnableRun flag (motors enabled)
        #   bit 3 -> 1 = Payload loaded, 0 = Payload unloaded - payload indicator
        #   bit 4 -> 1 = LIDAR ScanValid
        #   bit 5 -> 1 = Manual override
        # 4 byte SystemVoltage (float)  4 - battery level for control electronics [V]
        # 4 byte PowerVoltage (float)   8 - battery level for motors [V]
        self.flags, system_voltage, power_voltage = struct.unpack_from('<Iff', data, offset)
        self.lidar_valid = (self.flags & 0x10) == 0x10
        self.emergency_stop = (self.flags & 0x01) == 0x01

        self.voltage = [system_voltage, power_voltage]
        self.bus.publish('voltage', [int(v*100) for v in self.voltage])

        # skipped parsing of:
        # 4 byte SpeedM1 (float)       12 - normalized motor M1 (R) speed <-1.0 1.0>
        # 4 byte SpeedM2 (float)       16 - normalized motor M2 (L) speed <-1.0 1.0>
        motors = struct.unpack_from('<ff', data, offset + 12)

        # skipped parsing of:
        # 4 byte ActualDir (float)     20 - normalized direction for PID controller
        
        # 4 byte EncM1 (signed long)   24 - incremental encoders count for motor M1 (R) since last reset
        # 4 byte EncM2 (signed long)   28 - incremental encoders count for motor M2 (L) since last reset
        encoders = struct.unpack_from('<II', data, offset + 24)

        # skipped parsing of:
        # 1 byte GPS_Valid             32 - 1 = valid data from GPS module
        # 1 byte GPS_Fix               33 - 0 = no fix, 1 = 2D fix, 2 = 3D fix 
        # 4 byte GPS_UTCDate (ulong)   34 - GPS date in YYMMDD format
        # 4 byte GPS_UTCTime (ulong)   38 - GPS time in HHMMSS format
        # 4 byte GPS_Lat (ulong)       42 - format Lat * 1E7
        # 4 byte GPS_Lon (ulong)       46 - format Lon * 1E7
        # 4 byte GPS_Brg (float)       50 - GPS Bearing <0 .. 359> deg

        # 4 byte AHRS_q0 (float)       54 - Orientation Quaternion
        # 4 byte AHRS_q1 (float)       58 - 
        # 4 byte AHRS_q2 (float)       62 - 
        # 4 byte AHRS_q3 (float)       66 - 

        qw, qx, qy, qz = struct.unpack_from('<ffff', data, offset+54)
        orientation = qx, qy, qz, qw
        # identity quat points to north, we need it to point to east
        orientation = quaternion.multiply(orientation, [0, 0, 0.7071068, 0.7071068])
        # correct roll axis by 1.7 degrees
        self.orientation = quaternion.multiply(orientation, [0.0148348, 0, 0, 0.99989])
        self.bus.publish('orientation', list(self.orientation))

        q1, q2, q3, q0 = self.orientation  # quaternion
        #print(self.time, "{:.4f} {:.4f} {:.4f} {:.4f}".format(q0, q1, q2, q3))
        ax =  math.atan2(2*(q0*q1+q2*q3), 1-2*(q1*q1+q2*q2))
        ay =  math.asin(2*(q0*q2-q3*q1))
        az =  math.atan2(2*(q0*q3+q1*q2), 1-2*(q2*q2+q3*q3))
        # rotation Euler angles are yaw, pitch and roll
        #print(self.time, "{:.4f} {:.4f} {:.4f}".format(math.degrees(az), math.degrees(ay), math.degrees(ax)))
        self.bus.publish('rotation', [round(math.degrees(angle)*100) for angle in [az, ay, ax]])

        # 4 byte Yaw (float)           70 - Heading (Yaw) - machine orientation to magnetic north <0 .. 359> deg
        self.yaw = struct.unpack_from('<f', data, offset + 70)[0]
        #print(math.degrees(x), math.degrees(y), math.degrees(z), self.yaw)
        # 4 byte AccelX (float)        74
        # 4 byte AccelY (float)        78
        # 4 byte AccelZ (float)        82
        # 4 byte GyroX (float)         86
        # 4 byte GyroY (float)         90
        # 4 byte GyroZ (float)         94
        # 4 byte MagX (float)          98
        # 4 byte MagY (float)          102
        # 4 byte MagZ (float)          106
        # 4 byte SystemTick (ulong) 110  - Uptime in milisecond
        uptime = struct.unpack_from('<I', data, offset + 110)[0]
        if self.uptime is not None:
            uptime_diff = uptime - self.uptime
        self.uptime = uptime

        if self.last_encoders is not None:
            step = [sint32_diff(x, prev) for x, prev in zip(encoders, self.last_encoders)]
            self.publish('encoders', step)

            dist = ENC_SCALE * sum(step)/len(step)
            angle = ENC_SCALE * (step[0] - step[1])/WHEEL_DISTANCE
            x, y, heading = self.pose
            # advance robot by given distance and angle
            if abs(angle) < 0.0000001:  # EPS
                # Straight movement - a special case
                x += dist * math.cos(heading)
                y += dist * math.sin(heading)
                #Not needed: heading += angle
            else:
                # Arc
                r = dist / angle
                x += -r * math.sin(heading) + r * math.sin(heading + angle)
                y += +r * math.cos(heading) - r * math.cos(heading + angle)
                heading += angle # not normalized
            self.pose = (x, y, heading)
            self.send_pose()
        self.last_encoders = encoders

        # 4 byte LidarTimestamp (ulong) 114  - Value of SystemTick when lidar scan was received
        lidar_timestamp = struct.unpack_from('<I', data, offset + 114)[0]
        lidar_diff = lidar_timestamp - self.lidar_timestamp
        self.lidar_timestamp = lidar_timestamp
        if lidar_diff > 150 and self.lidar_valid:
            print(self.time, "lidar invalid:", lidar_diff)
            self.lidar_valid = False
        if lidar_diff != 0 and self.lidar_valid:
            # laser
            # 480 byte Lidar_Scan (ushort) 118 - 239 two-bytes distances from Lidar <0 .. 65535> in [cm]
            # Scan is whole 360 deg with resolution 1.5 deg
            scan = struct.unpack_from('<' + 'H'*239, data, offset + 118)  # TODO should be 240

            # restrict scan only to 270 degrees - cut 1/8th on both sides
#            scan = scan[30:-30]
#            zero_sides = 20
            scan = [10 * d for d in reversed(scan)]  # scale to millimeters
#            scan[:zero_sides] = [0]*zero_sides
#            scan[-zero_sides:] = [0]*zero_sides
            self.publish('scan', scan)

    def run(self):
        try:
            self.publish('raw', self.query_version())
            while True:
                dt, channel, data = self.listen()
                self.time = dt
                if channel == 'raw':
                    self._buf += data
                    packet = self.get_packet()
                    if packet is not None:
                        if len(packet) < 256:  # TODO cmd value
                            print(packet)
                        else:
                            prev = self.flags
                            self.parse_packet(packet)
                            if prev != self.flags:
                                print(self.time, 'Flags:', hex(self.flags))
                        self.publish('raw', self.create_packet())
                if channel == 'desired_speed':
                    self.desired_speed, self.desired_angular_speed = data[0]/1000.0, math.radians(data[1]/100.0)                    
                    if abs(self.desired_speed) < 0.2 and abs(self.desired_angular_speed) > 0.2:
                        if self.speeds.__name__ != "oscilate":
                            self.speeds = self.oscilate()
                    else:
                        if self.speeds.__name__ == "oscilate":
                            self.speeds = self.plain_speeds()
                    self.cmd_flags |= 0x02  # PWM ON
#                    if data == [0, 0]:
#                        print("TURN OFF")
#                        self.cmd_flags = 0x00  # turn everything OFF (hack for now)

        except BusShutdownException:
            pass

# vim: expandtab sw=4 ts=4



from threading import Thread

from osgar.bus import BusShutdownException
from osgar.lib import quaternion


class Offseter:

    def __init__(self, config, bus):
        bus.register("pose3d")
        self.thread = Thread(target=self.run)
        #self.thread.name = bus.name
        self.bus = bus
        self.is_alive = self.thread.is_alive
        self.init_quat = None  # unknown rotation

    def start(self):
        self.thread.start()

    def join(self, timeout=None):
        self.thread.join(timeout=timeout)

    def _run(self):
        origin = None
        xyz = None
        while None in (origin, xyz):
            dt, channel, data = self.bus.listen()
            if channel == "origin":
                if len(data) == 8:
                    robot_name, x, y, z, qa, qb, qc, qd = data
                    origin = [x, y, z]
                    if self.init_quat is None:
                        self.init_quat = (qa, qb, qc, qd)
            elif channel == "pose3d":
                xyz, orientation = data
            else:
                raise RuntimeError(f"unknown channel '{channel}'")

        offset = [o - p for o,p in zip(origin, xyz)]

        while True:
            dt, channel, data = self.bus.listen()
            if channel == "origin":
                pass  # do not update offset estimate
            elif channel == "pose3d":
                xyz, orientation = data
                xyz2 = quaternion.rotate_vector(xyz, self.init_quat)
                xyz_offset = [o + p for o,p in zip(offset, xyz2)]
                orientation_offset = quaternion.multiply(orientation, self.init_quat)
                self.bus.publish("pose3d", [xyz_offset, orientation_offset])
            else:
                raise RuntimeError(f"unknown channel '{channel}'")

    def run(self):
        try:
            self._run()
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()


def main():
    import argparse
    import datetime
    from unittest.mock import MagicMock
    from osgar.logger import LogReader, lookup_stream_id
    from osgar.lib.serialize import deserialize
    from osgar.bus import Bus

    parser = argparse.ArgumentParser()
    parser.add_argument('logfile', help='filepath')
    args = parser.parse_args()

    origin_stream_id = lookup_stream_id(args.logfile, "rosmsg.origin")
    pose3d_stream_id = lookup_stream_id(args.logfile, "rospy.pose3d")
    channel = { origin_stream_id:"origin", pose3d_stream_id: "pose3d" }
    print(channel)

    bus = Bus(MagicMock(write=MagicMock(return_value=datetime.timedelta())))
    tester = bus.handle("tester")
    tester.register("pose3d", "origin")
    def aa():
        while True:
            dt, channel, data = tester.listen()
            print(channel, data)
    tester_thread = Thread(target=aa, name='tester')

    offseter_handle = bus.handle("offseter")
    offseter_node = Offseter({}, offseter_handle)

    bus.connect("tester.pose3d", "offseter.pose3d")
    bus.connect("tester.origin", "offseter.origin")
    bus.connect("offseter.pose3d", "testerer.pose3d")

    offseter_node.start()
    tester_thread.start()

    try:
        with LogReader(args.logfile, only_stream_id=[origin_stream_id, pose3d_stream_id]) as logreader:
            for time, stream, data in logreader:
                data = deserialize(data)
                tester.publish(channel[stream], data)
    except Exception as e:
        print(e)
        offseter_node.request_stop()
        tester.request_shutdown()
        offseter_node.join()
        tester.join()


if __name__ == "__main__":
    main()



"""
  Wrapper & timestamper of input/out over socket
"""

import socket
import urllib.request
from threading import Thread

from osgar.logger import LogWriter
from osgar.bus import BusShutdownException


class LogSocket:
    def __init__(self, socket, config, bus):
        bus.register('raw')
        self.verbose = False
        self.socket = socket

        self.input_thread = Thread(target=self.run_input, daemon=True)
        self.output_thread = Thread(target=self.run_output, daemon=True)

        host = config.get('host')
        port = config.get('port')
        self.pair = (host, port)  # (None, None) for unknown address
        if 'timeout' in config:
            self.socket.settimeout(config['timeout'])
        self.bufsize = config.get('bufsize', 1024)

        self.bus = bus

    def _send(self, data):
        raise NotImplementedError()

    def start(self):
        self.input_thread.start()
        self.output_thread.start()

    def join(self, timeout=None):
        self.input_thread.join(timeout=timeout)
        self.output_thread.join(timeout=timeout)

    def run_input(self):
        while self.bus.is_alive():
            try:
                data = self.socket.recv(self.bufsize)
                if len(data) > 0:
                    self.bus.publish('raw', data)
            except socket.timeout:
                pass

    def run_output(self):
        try:
            while True:
                __, channel, data = self.bus.listen()
                if channel == 'raw':
                    self._send(data)
                else:
                    assert False, channel  # unsupported channel
 
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()


class LogTCPBase(LogSocket):
    """
      TCP base class for different use cases
    """
    def __init__(self, config, bus):
        soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        super().__init__(soc, config, bus)

    def _send(self, data):
        self.socket.send(data)


class LogTCPStaticIP(LogTCPBase):
    """
      TCP driver for existing static IP (i.e. SICK LIDAR)
    """
    def __init__(self, config, bus):
        super().__init__(config, bus)
        try:
            self.socket.connect(self.pair)
        except socket.timeout as e:
            print('Timeout', self.pair)
            raise


class LogTCPDynamicIP(LogTCPBase):
    """
      TCP driver for dynamic previously unknown address
         (ROS proxy for subscribers)
    """
    def start(self):
        # on start the address is unknown - it will be received by "output_thread"
        self.output_thread.start()

    def join(self, timeout=None):
        # the "input_thread" is triggered by "output_thread" so make sure
        # that "output_thread" is finished and cannot cause race condition
        self.output_thread.join(timeout=timeout)
        try:
            self.input_thread.join(timeout=timeout)
        except RuntimeError:
            pass

    def run_output(self):
        try:
            while True:
                __, channel, data = self.bus.listen()
                if channel == 'raw':
                    self._send(data)
                elif channel == 'addr':
                    self.pair = tuple(data)
                    self.socket.connect(self.pair)
                    if not self.input_thread.is_alive():
                        self.input_thread.start()
                else:
                    assert False, channel  # unsupported channel
 
        except BusShutdownException:
            pass


class LogTCPServer(LogTCPBase):
    """
      TCP driver for server side - prepare connection and wait
      for others to connect (ROS proxy for publishers)
    """

    def __init__(self, config, bus):
        super().__init__(config, bus)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.pair)
        self.timeout = config.get('timeout')

    def run_input(self):
        if self.verbose:
            print("Waiting ...")
        self.socket.listen(1)
        if self.verbose:
            print("end of listen")
        while self.bus.is_alive():
            try:
                self.socket, addr = self.socket.accept()
                if self.verbose:
                    print('Connected by', addr)
                if self.timeout is not None:
                    self.socket.settimeout(self.timeout)
                super().run_input()
                break
            except socket.timeout:
                pass


class LogUDP(LogSocket):
    def __init__(self, config, bus):
        soc = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        LogSocket.__init__(self, soc, config, bus)
        try:
            self.socket.bind(('', self.pair[1]))
        except OSError as e:
            if e.errno == 98: # [Errno 98] Address already in use
                self.socket.connect(('', self.pair[1]))

    def _send(self, data):
        self.socket.sendto(data, self.pair)


class LogHTTP:
    def __init__(self, config, bus):
        bus.register('raw')
        self.input_thread = Thread(target=self.run_input, daemon=True)

        self.url = config['url']
        self.sleep = config.get('sleep', None)
        self.bus = bus

    def start(self):
        self.input_thread.start()

    def join(self, timeout=None):
        self.input_thread.join(timeout=timeout)

    def run_input(self):
        while self.bus.is_alive():
            try:
                # https://github.com/mesonbuild/meson/issues/4087
                # without timeout the call can hang the process forever
                with urllib.request.urlopen(self.url, timeout=0.5) as f:
                    data = f.read()
                if len(data) > 0:
                    self.bus.publish('raw', data)
            except socket.timeout:
                pass
            if self.sleep is not None:
                self.bus.sleep(self.sleep)

    def request_stop(self):
        self.bus.shutdown()


if __name__ == "__main__":
    import time
    from osgar.bus import Bus

    config = {'host':'localhost', 'port': 8001, 'timeout': 1.0}
    with LogWriter(prefix='test-tcp-') as log:
        bus = Bus(log)
        device = LogTCPStaticIP(config, bus=bus.handle('tcp'))
        device.start()
        time.sleep(2)
        device.request_stop()
        device.join()

# vim: expandtab sw=4 ts=4



"""
  Parse 3DM-GX5-25 data
"""
# www.microstrain.com/sites/default/files/3dm-gx5-25_dcp_manual_8500-0065.pdf
import struct
import math
import datetime

from osgar.node import Node


def verify_checksum(packet):
    # 16-bit Fletcher Checksum Algorithm
    ch1, ch2 = 0, 0
    for b in packet[:-2]:
        ch1 += b
        ch2 += ch1
    ret = ((ch1 << 8) & 0xFF00) + (ch2 & 0xFF)
    return packet[-2] == (ch1 & 0xFF) and packet[-1] == (ch2 & 0xFF)


def get_packet(buf):
    try:
        i = buf.index(b'\x75\x65')
    except ValueError:
        return None, buf
    try:
        packet_size = buf[i + 3] + 4 + 2  # + header size + checksum
        if i + packet_size > len(buf):
            return None, buf
    except IndexError:
        return None, buf
    packet = buf[i:i + packet_size]
    if not verify_checksum(packet):
        print('LORD Checksum error:', packet.hex())
        packet = None
    return packet, buf[i + packet_size:]


def parse_packet(packet, verbose=False):
    assert packet.startswith(b'\x75\x65'), packet.hex()
    assert len(packet) > 3, len(packet)
    packet_size = packet[3] + 4 + 2 # + header size + checksum
    assert len(packet) == packet_size, (len(packet), packet_size)
    desc = packet[2]

    # IMU Data Set (0x80)
    assert desc in [0x80, 0x81, 0x82], hex(desc)
    i = 4
    acc, gyro, euler, quat, gps = None, None, None, None, None
    lat, lon, sat_num, gps_time = None, None, None, None
    while i < packet_size - 2:
        field_length = packet[i]
        cmd = packet[i + 1]

        if desc == 0x80 and cmd == 0x04:
            # Scaled Accelerometer Vector
            assert field_length == 14, field_length
            acc = struct.unpack_from('>fff', packet, i + 2)
            if verbose:
                print('acc', acc)
        elif desc == 0x80 and cmd == 0x05:
            # Scaled Gyro Vector (0x80, 0x05)
            assert field_length == 14, field_length
            gyro = struct.unpack_from('>fff', packet, i + 2)
            if verbose:
                print('gyro', gyro)
        elif desc == 0x80 and cmd == 0x06:
            # Scaled Magnetometer Vector (0x80, 0x06)
            assert field_length == 14, field_length
            mag = struct.unpack_from('>fff', packet, i + 2)
            if verbose:
                print('mag', mag)
        elif desc == 0x80 and cmd == 0x0C:
            # CF Euler Angles (0x80, 0x0C)
            # This value is produced by the Complementary Filter fusion algorithm.
            pass

        # GPS related messages
        elif desc == 0x81 and cmd == 0x03:
            # LLH Position (0x81, 0x03)
            assert field_length == 44, field_length
            row = struct.unpack_from('>ddddffH', packet, i + 2)
            lat, lon, height_elipsoid, height_msl, hori, vert, flags = row
            if verbose:
                print('GPS', (lat, lon))
        elif desc == 0x81 and cmd == 0x08:
            # UTC Time (0x81, 0x08)
            assert field_length == 15, field_length
            msg = struct.unpack_from('>HBBBBBIH', packet, i + 2)
            year, month, day, hour, minute, second, millisecond, flags = msg
            gps_time = datetime.datetime(year, month, day, hour, minute, second, millisecond)
            if verbose:
                print('UTCTime', year, month, day, hour, minute, second, millisecond, flags)
        elif desc == 0x81 and cmd == 0x0b:
            # GNSS Fix Information (0x81, 0x0B)
            assert field_length == 8, field_length
            msg = struct.unpack_from('>BBHH', packet, i + 2)
            fix_type, sat_num, fix_flags, valid = msg
            if verbose:
                print('GPSFix', msg)

        elif desc == 0x82 and cmd == 0x03:
            # Orientation, Quaternion (0x82, 0x03)
            q0, q1, q2, q3, valid = struct.unpack_from('>ffffH', packet, i + 2)
            if valid == 0x1:
                quat = q0, q1, q2, q3
            if verbose:
                print('quaterion', q0, q1, q2, q3, valid)
        elif desc == 0x82 and cmd == 0x05:
            # Orientation, Euler Angles (0x82, 0x05)
            roll, pitch, yaw, valid = struct.unpack_from('>fffH', packet, i + 2)
            if valid == 0x1:
#                euler = yaw, pitch, roll
#                euler = math.pi - yaw, pitch, roll  # hacked, guessed
                euler = math.pi - yaw, pitch, math.pi - roll  # hacked2, guessed for K2
            if verbose:
                print('euler', roll, pitch, yaw, valid)
        elif desc == 0x82 and cmd == 0x0A:
            # Attitude Uncertainty, Euler Angles (0x82, 0x0A)
            roll, pitch, yaw, valid = struct.unpack_from('>fffH', packet, i + 2)
            if verbose:
                print('euler-uncertain', roll, pitch, yaw, valid)
        elif desc == 0x82 and cmd == 0x0D:
            # Linear Acceleration (0x82, 0x0D)
            # Filter-Compensated Linear Acceleration Data (gravity vector removed)
            x, y, z, valid = struct.unpack_from('>fffH', packet, i + 2)
            if verbose:
                print('lin', x, y, z, valid)
        elif desc == 0x82 and cmd == 0x0E:
            # Compensated Angular Rate (0x82, 0x0E)
            x, y, z, valid = struct.unpack_from('>fffH', packet, i + 2)
            if verbose:
                print('ang', x, y, z, valid)
        elif desc == 0x82 and cmd == 0x14:
            # Heading Update Source information expressed in the sensor frame.
            heading, uncertainty, source, valid = struct.unpack_from('>ffHH', packet, i + 2)
            if verbose:
                print('heading', heading, uncertainty, source, valid)
        else:
            assert False, (hex(desc), hex(cmd))

        i += field_length

    if lat is not None:
        gps = (lat, lon), sat_num, gps_time

    return acc, gyro, euler, quat, gps


class LordIMU(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('orientation', 'rotation', 'gps_position')
        self._buf = b''
        self.raw = None  # not automatically defined yet
        self.verbose = False

    def update(self):
        channel = super().update()
        assert channel == 'raw', channel
        self._buf += self.raw 
        while True:
            packet, self._buf = get_packet(self._buf)
            if packet is None:
                break
            acc, gyro, euler, quat, gps = parse_packet(packet, self.verbose)
            if euler is not None:
                yaw, pitch, roll = euler
                # This is a three component vector containing the Roll, Pitch and Yaw angles 
                # in radians.
                self.publish('rotation', 
                        [int(round(math.degrees(x) * 100)) for x in [yaw, pitch, roll]])
            if quat is not None:
                self.publish('orientation', list(quat))
            if gps is not None:
                # TODO check for GPS fix
                lat, lon = gps[0]
                x = int(lon * 3600000)  # deg to milisec
                y = int(lat * 3600000)
                self.publish('gps_position', [x, y])

# vim: expandtab sw=4 ts=4



"""
Calibrate camera using fisheye model from opencv.

https://docs.opencv.org/trunk/db/d58/group__calib3d__fisheye.html#details

Collect images to `images/capture*.jpg` by pressing "S".
If precaptured images are found, calibration is run directly.
It is expected that chessboard calibration pattern 9x6 is on the images.
Suitable pattern can be generated using
https://calib.io/pages/camera-calibration-pattern-generator

https://stackoverflow.com/a/50876130/3185929
https://github.com/njanirudh/Aruco_Tracker/blob/master/camera_calibration.py
"""

import cv2
import glob
import numpy as np

PATTERN = (8,5)

def capture(camera=0):
    cap = cv2.VideoCapture(camera)
    if not cap.isOpened():
        return

    cap.set(cv2.CAP_PROP_FRAME_WIDTH, 1280)
    cap.set(cv2.CAP_PROP_FRAME_HEIGHT, 720)
    cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*"MJPG"))

    count = 0
    while True:
        status, img = cap.read()
        if status == False:
            continue
        gray = cv2.cvtColor(img,cv2.COLOR_BGR2GRAY)
        gray = cv2.resize(gray, (img.shape[1]//2, img.shape[0]//2))
        found, corners = cv2.findChessboardCorners(gray, PATTERN, cv2.CALIB_CB_ADAPTIVE_THRESH+cv2.CALIB_CB_FAST_CHECK+cv2.CALIB_CB_NORMALIZE_IMAGE)
        img_copy = img.copy()
        if found:
            corners *= 2
            cv2.drawChessboardCorners(img, PATTERN, corners, found)
        cv2.imshow('img', img)
        key = cv2.waitKey(1)
        if key == 27:
            return
        if key == ord('s'):
            if found:
                filename = 'capture{:02d}.jpg'.format(count)
                print(filename)
                cv2.imwrite('images/'+filename, img_copy)
                count += 1


def calibrate(images):
    subpix_criteria = cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.1
    calibration_flags = cv2.fisheye.CALIB_RECOMPUTE_EXTRINSIC + cv2.fisheye.CALIB_CHECK_COND + cv2.fisheye.CALIB_FIX_SKEW
    # prepare object points, like (0,0,0), (1,0,0), (2,0,0) ....,(PATTERN[0],PATTERN[1],0)
    # but for some crazy reasons it needs to have the shape (1, x, 3)
    objp = np.zeros((1, PATTERN[0]*PATTERN[1],3), np.float32)
    objp[0,:,:2] = np.mgrid[0:PATTERN[0], 0:PATTERN[1]].T.reshape(-1,2)
    objpoints = [] # 3d point in real world space
    imgpoints = [] # 2d points in image plane.
    for image in sorted(images):
        gray = cv2.imread(image, cv2.IMREAD_GRAYSCALE)
        found, corners = cv2.findChessboardCorners(gray, PATTERN, cv2.CALIB_CB_ADAPTIVE_THRESH+cv2.CALIB_CB_FAST_CHECK+cv2.CALIB_CB_NORMALIZE_IMAGE)
        print(image, found)
        if found:
            objpoints.append(objp)
            cv2.cornerSubPix(gray, corners, (3,3), (-1,-1), subpix_criteria)
            imgpoints.append(corners)
    K = np.zeros((3, 3))
    D = np.zeros((4, 1))
    rms, _, _, _, _ = cv2.fisheye.calibrate(objpoints, imgpoints, gray.shape[::-1], K, D,
        flags=calibration_flags, criteria=(cv2.TERM_CRITERIA_EPS+cv2.TERM_CRITERIA_MAX_ITER, 30, 1e-6))

    return len(objpoints), rms, gray.shape[::-1], K, D


def fish2perspective(images, K, D, DIM):
    NK = K.copy()
    NK[0,0] /= 2
    NK[1,1] /= 2

    map1, map2 = cv2.fisheye.initUndistortRectifyMap(K, D, np.eye(3), NK, DIM, cv2.CV_16SC2)
    for fname in sorted(images):
        img = cv2.imread(fname)
        undistorted_img = cv2.remap(img, map1, map2, interpolation=cv2.INTER_LINEAR, borderMode=cv2.BORDER_CONSTANT)
        cv2.imshow('a', undistorted_img)
        print(fname)
        if cv2.waitKey(0) == 27:
            return


def main():
    import argparse

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('--camera', help='camera id', type=int, default=0)
    args = parser.parse_args()

    images = glob.glob('images/capture*.jpg')
    if len(images) == 0:
        capture(args.camera)
    images = glob.glob('images/capture*.jpg')
    if len(images) == 0:
        return
    N, rms, DIM, K, D = calibrate(images)
    print("Found " + str(N) + " valid images for calibration")
    print("RMS = " + str(rms))
    print("DIM = " + str(DIM))
    print("K = np.array(" + str(K.tolist()) + ")")
    print("D = np.array(" + str(D.tolist()) + ")")

    fish2perspective(images, K, D, DIM)


if __name__ == "__main__":
    main()



import unittest
import math

from osgar.explore import tangent_circle, follow_wall_angle


class ExploreTest(unittest.TestCase):

    def test_tangent_circle(self):
        angle = tangent_circle(dist=2.0, radius=1.0)
        self.assertAlmostEqual(angle, math.radians(30))
#        self.assertIsNone(angle)  # too far

        angle = tangent_circle(dist=2.0, radius=3.0)
#        self.assertIsNone(angle)  # no tangent
        self.assertAlmostEqual(angle, math.radians(100))

        angle = tangent_circle(dist=math.sqrt(2), radius=1.0)
        self.assertAlmostEqual(angle, math.radians(45))

    def test_follow_wall_angle(self):
        # nothing close -> no command
        scan = [0] * 271  # 1deg resolution
        cmd = follow_wall_angle(scan, radius=1.0, right_wall=True)
        self.assertAlmostEqual(cmd, math.radians(-20))

        # parallel wall at 1m on the right
        scan = []
        for i in range(-135, 135+1):
            # ax + by + c = 0
            #  y = -1
            # ray: (t*cos(angle), t*sin(angle))
            #  t*sin(angle) = -1
            #  t = -1/sin(angle)
            if i < 0:
                dist_mm = -round(1000/math.sin(math.radians(i)))
            else:
                dist_mm = 0  # no reflection
            scan.append(dist_mm)
        cmd = follow_wall_angle(scan, radius=1.0, right_wall=True)
        self.assertIsNotNone(cmd)
        # angle should be 0.0, but due to resolution it corresponds to 1deg
        self.assertLess(abs(math.degrees(cmd)), 1.5)

        # move towards wall
        cmd = follow_wall_angle(scan, radius=0.7, right_wall=True)
        self.assertIsNotNone(cmd)
        self.assertLess(math.degrees(cmd), 0)  # used to be -45, now only slightly move right

        # move from wall
        cmd = follow_wall_angle(scan, radius=1.3, right_wall=True)
        self.assertIsNotNone(cmd)
        self.assertGreater(math.degrees(cmd), 0)  # just move from the wall (used to be 8deg)

# vim: expandtab sw=4 ts=4



"""
  Wrapper for USB communication
"""

from threading import Thread
import time

# pyusb
import usb.core
import usb.util

from osgar.bus import BusShutdownException


class LogUSB:
    def __init__(self, config, bus):
        self.thread = Thread(target=self.run, daemon=True)

        # TODO make configurable for other devices then SICK lidar TIM310
        self.dev = usb.core.find(idVendor=0x19A2, idProduct=0x5001)
        for i in range(10):
            try:
                self.dev.set_configuration()
                break
            except:
                print("LaserUSB - init ERROR", i)
                time.sleep(0.1)

        self.bus = bus

    def start(self):
        self.thread.start()

    def join(self, timeout=None):
        self.thread.join(timeout=timeout)

    def run(self):
        try:
            while True:
                __, __, data = self.bus.listen()
                try:
                    self.dev.write(2|usb.ENDPOINT_OUT, data, 0)
                    data = self.dev.read(1|usb.ENDPOINT_IN, 65535, timeout = 100)
                    self.bus.publish('raw', bytes(data))
                except usb.core.USBError as error:
                    print(error)
                    self.bus.publish('raw', bytes())
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()

# vim: expandtab sw=4 ts=4



"""
  OSGAR Gas detector
"""
# This is the simplest version taken from subt/artifacts.py - i.e. use reference position as
# the first moment when the robot detects gas. More advanced/precise implementations can take the map
# or scan and report the entering doors as defined in the rules.

from osgar.node import Node
from subt.artifacts import GAS


class ArtifactGasDetector(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register("artf")

    def on_gas_detected(self, data):
        # [type, [rel_x, rel_y, rel_z]]
        if data:
            self.publish('artf', [GAS, [0, 0, 0]])

    def update(self):
        channel = super().update()
        handler = getattr(self, "on_" + channel, None)
        if handler is not None:
            handler(getattr(self, channel))
        else:
            assert False, channel  # not supported channel

# vim: expandtab sw=4 ts=4



"""
  Convert depth image to "lidar" scan
"""
import math

import cv2
import numpy as np

from osgar.node import Node
from osgar.bus import BusShutdownException
from osgar.lib.depth import depth2dist, DepthParams
from osgar.lib.mathex import normalizeAnglePIPI


FRAC = math.tan(math.radians(60)) / 360


def vertical_scan(depth, column):
    # return two arrays x and y
    arr = np.array(depth[:, column], np.int32)
    x = arr
    a = FRAC * (np.arange(len(arr), 0, -1) - 180)
    y = x * a + 560
    return x, y


def monotize(arr):
    """change function to monotonous increasing"""
    best = arr[0]
    for i in range(len(arr)):  # TODO optimize
        best = max(best, arr[i])
        arr[i] = best
    return arr


def find_step(arr):
    """input is uniform array of values"""
    INDEX_STEP = 10  # so for 10cm step there is 60mm height difference -> ~30deg
    THRESHOLD = 60
    diff = arr[INDEX_STEP:] - arr[:-INDEX_STEP]
    mask = diff > THRESHOLD
    index = np.argmax(mask)
    if mask[index]:
        return index + INDEX_STEP
    return None


def vertical_step(depth, column=320):
    """Detect nearest obstacle for vertical line"""
    MAX_SIZE = 250  # in cm, 1cm resolution
    x_arr, y_arr = vertical_scan(depth, column)
    x_arr //= 10  # reduction to cm
    d = np.zeros(MAX_SIZE, np.int32)
    for x, y in zip(x_arr, y_arr):
        if 0 <= x < MAX_SIZE:
            d[x] = max(d[x], y)
    ret = find_step(monotize(d))
    if ret is None:
        return None
    return ret * 10  # return back readings in mm


def is_on_line(a, b, c, z=300):
    "Is point C at most `z` milimeters far from a line defined by points A and B?"
    ax, ay = a
    bx, by = b
    cx, cy = c
    p = -(by - ay)
    q = bx - ax
    r = -(p * ax + q * ay)
    d = (p * cx + q * cy + r) / np.hypot(p, q)
    return abs(d) < z


def adjust_scan(scan, depth_scan, depth_params):

    # Lidar scan starts with right-most direction. Let's make the
    # virtual scan based on depth camera do the same.
    depth_scan = np.asarray(depth_scan[::-1])

    lidar_scan = np.asarray(scan)

    camera_fov = 2 * np.arctan2(depth_params.camw / 2.0, depth_params.fx)
    lidar_fov = depth_params.lidar_fov

    depth_density = depth_scan.shape[0] / camera_fov
    lidar_density = lidar_scan.shape[0] / lidar_fov
    assert depth_density > lidar_density
    depth_scan = cv2.resize(
        depth_scan.reshape((1, -1)),
        (int(camera_fov * lidar_density), 1),
        interpolation=cv2.INTER_NEAREST).reshape((-1,))
    alignment_start = (lidar_scan.shape[0] - depth_scan.shape[0]) // 2
    alignment_tail = lidar_scan.shape[0] - depth_scan.shape[0] - alignment_start

    # Part of the lidar scan that overlaps with depth camera.
    lidar_overlap = lidar_scan[alignment_start:-alignment_tail]

    DISAGREEMENT_MM_LIMIT = 600
    LIDAR_NO_MEASUREMENT = 0
    DEPTH_NO_MEASUREMENT = 0
    def disagreement(idx):
        return lidar_overlap[idx] != LIDAR_NO_MEASUREMENT and (
                depth_scan[idx] == DEPTH_NO_MEASUREMENT or
                depth_scan[idx] - lidar_overlap[idx] > DISAGREEMENT_MM_LIMIT)
    # We check multiple consecutive directions to reduce noise.
    disagreement_right = disagreement(0) and disagreement(1) and disagreement(2)
    disagreement_left = disagreement(-1) and disagreement(-2) and disagreement(-3)

    if disagreement_right:
        # Taking measurements 2 indices away to lower impact of noise.
        d0, d1 = lidar_scan[alignment_start], lidar_scan[alignment_start + 2]
        phi0 = -lidar_fov / 2 + alignment_start / lidar_density
        phi1 = phi0 + 2 / lidar_density
        a = d0 * np.cos(phi0), d0 * np.sin(phi0)
        b = d1 * np.cos(phi1), d1 * np.sin(phi1)
        for i in range(alignment_start):
            d2 = lidar_scan[i]
            if d2 == LIDAR_NO_MEASUREMENT or d2 < depth_params.lidar_trusted_zone:
                continue
            phi2 = -lidar_fov / 2 + i / lidar_density
            c = d2 * np.cos(phi2), d2 * np.sin(phi2)
            if is_on_line(a, b, c):
                lidar_scan[i] = LIDAR_NO_MEASUREMENT
    if disagreement_left:
        # Taking measurements 2 indices away to lower impact of noise.
        aidx = -alignment_tail - 1
        bidx = -alignment_tail - 1 - 2
        d0, d1 = lidar_scan[aidx], lidar_scan[bidx]
        phi0 = lidar_fov / 2 + aidx / lidar_density
        phi1 = phi0 - 2 / lidar_density
        a = d0 * np.cos(phi0), d0 * np.sin(phi0)
        b = d1 * np.cos(phi1), d1 * np.sin(phi1)
        lidar_len = lidar_scan.shape[0]
        for i in range(lidar_len - alignment_tail, lidar_len):
            d2 = lidar_scan[i]
            if d2 == LIDAR_NO_MEASUREMENT or d2 < depth_params.lidar_trusted_zone:
                continue
            phi2 = -lidar_fov / 2 + i / lidar_density
            c = d2 * np.cos(phi2), d2 * np.sin(phi2)
            if is_on_line(a, b, c):
                lidar_scan[i] = LIDAR_NO_MEASUREMENT

    depth_scan[depth_scan == DEPTH_NO_MEASUREMENT] = LIDAR_NO_MEASUREMENT
    new_scan = lidar_scan
    # Up to depth_params.lidar_trusted_zone, we trust the lidar measurements,
    # so we take a minimum between lidar- and depth-based distance to obstacle.
    # Beyond that distance, we no longer trust lidar, because it could, for
    # example, be pointing to ground, and we only take the depth-based estimate.
    new_scan[alignment_start:-alignment_tail] = np.where(
            np.logical_and(
                lidar_overlap != LIDAR_NO_MEASUREMENT,
                lidar_overlap < depth_params.lidar_trusted_zone),
            np.minimum(lidar_overlap,
                   # If we have a trusted lidar measurement, we shouldn't
                   # override it with "no depth measurement", even if that is
                   # technically a lower value.
                   np.where(
                       depth_scan == DEPTH_NO_MEASUREMENT,
                       np.inf,
                       depth_scan)),
            depth_scan)
    return new_scan

class DepthToScan(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register("scan")
        self.depth = None  # initialize inputs
        self.scan = None
        self.verbose = False
        self.scale = np.array([1/math.cos(math.radians(30*(i-80)/80)) for i in range(160)])
        self.yaw, self.pitch, self.roll = None, None, None  # unknown values
        self.depth_params = DepthParams(**config.get('depth_params', {}))

    def update(self):
        channel = super().update()
        assert channel in ["depth", "scan", "rot"], channel

        if channel == 'depth':
            pass
        elif channel == 'scan':
            if self.depth is None:
                self.publish('scan', self.scan)
                return channel  # when no depth data are available ...
            depth_scan = depth2dist(self.depth, self.depth_params, self.pitch, self.roll)
            new_scan = adjust_scan(self.scan, depth_scan, self.depth_params)
            self.publish('scan', new_scan.tolist())
        elif channel == 'rot':
            self.yaw, self.pitch, self.roll = [normalizeAnglePIPI(math.radians(x/100)) for x in self.rot]
        else:
            assert False, channel  # unsupported channel

        return channel



if __name__ == '__main__':
    import argparse
    import matplotlib.pyplot as plt

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('filename', help='NPZ file with depth data')
    parser.add_argument('-v', '--verbose', help='verbose mode', action='store_true')
    parser.add_argument('-d', '--draw', help='draw graph(s)', action='store_true')
    args = parser.parse_args()

    with np.load(args.filename) as f:
        depth = f['depth']

    if args.draw:
        for i in range(0, 640, 159):
            x, y = vertical_scan(depth, i)
            plt.plot(x, y, 'o-', linewidth=2)

            d =  vertical_step(depth, i)
            plt.plot([d, d], [0, 200], '-', linewidth=3)

        plt.axes().set_aspect('equal', 'datalim')
        plt.show()

# vim: expandtab sw=4 ts=4



import unittest
import math
from datetime import timedelta

from osgar.lib.virtual_bumper import VirtualBumper


class VirtualBumperTest(unittest.TestCase):    
    def test_usage(self):
        vb = VirtualBumper(timedelta(seconds=2.5), dist_radius_limit=0.1)
        t0 = timedelta(seconds=12)
        t_step = timedelta(seconds=1)
        pose = [1.0, 2.0, math.pi]
        vb.update_desired_speed(0.5, 0.0)
        self.assertFalse(vb.collision())
        vb.update_pose(t0, pose)
        vb.update_pose(t0 + t_step, pose)
        vb.update_pose(t0 + 2*t_step, pose)
        vb.update_pose(t0 + 3*t_step, pose)
        self.assertTrue(vb.collision())

        vb.update_desired_speed(0.0, 0.0)
        self.assertFalse(vb.collision())

    def test_motion(self):
        vb = VirtualBumper(timedelta(seconds=2), dist_radius_limit=0.1)
        t0 = timedelta(seconds=0)
        t_step = timedelta(seconds=1)
        pose = [0.0, 0.0, 0.0]
        vb.update_pose(t0, pose)
        vb.update_desired_speed(0.5, 0.0)
        self.assertFalse(vb.collision())

        pose = [0.1, 0.0, 0.0]
        vb.update_pose(t0 + t_step, pose)
        pose = [0.2, 0.0, 0.0]
        vb.update_pose(t0 + 2*t_step, pose)
        self.assertFalse(vb.collision())

    def test_real_data(self):
        # taken from Eduro
        poses = [
            (7.213, -1.344, -1.55439023182615),
            (7.213, -1.344, -1.5563100940033436),
            (7.213, -1.344, -1.556833692778942),
            (7.213, -1.344, -1.555611962302546),
            (7.213, -1.344, -1.556833692778942),
            (7.213, -1.344, -1.556833692778942),
            (7.213, -1.344, -1.55439023182615),
            (7.213, -1.344, -1.5538666330505517),
            (7.213, -1.344, -1.5580554232553379),
            (7.213, -1.344, -1.55439023182615),
            (7.213, -1.344, -1.556833692778942),
            (7.213, -1.344, -1.556833692778942)]
        vb = VirtualBumper(timedelta(seconds=2), dist_radius_limit=0.1)
        t0 = timedelta(seconds=0)
        t_step = timedelta(seconds=0.5)
        vb.update_desired_speed(0.5, 0.0)
        for i, pose in enumerate(poses):
            vb.update_pose(t0 + i*t_step, pose)
        self.assertTrue(vb.collision())

    def test_false_trigger_via_lora_pause(self):
        # bug shown during LoRa command "Pause"
        vb = VirtualBumper(timedelta(seconds=2), dist_radius_limit=0.1)
        vb.update_desired_speed(0.5, 0.0)
        vb.update_pose(timedelta(seconds=2.479321), (0.0, 0.0, 0.0))
        vb.update_pose(timedelta(seconds=27.612698), (2.503, -0.013, -0.00977))
        vb.update_desired_speed(0.0, 0.0)
        vb.update_pose(timedelta(seconds=27.767734), (2.507, -0.013, -0.0095993))
        vb.update_pose(timedelta(seconds=30.458526), (2.525, -0.014, -0.0076794487))
        vb.update_pose(timedelta(seconds=37.808329), (2.525, -0.014, -0.0089011791851))
        self.assertFalse(vb.collision())

    def test_turn_in_place(self):
        # first current version without checking angle
        vb = VirtualBumper(timedelta(seconds=2), dist_radius_limit=0.1)
        vb.update_desired_speed(0.0, math.pi/2)
        self.assertTrue(vb.should_be_moving)
        vb.update_pose(timedelta(seconds=0), (0.0, 0.0, 0.0))
        vb.update_pose(timedelta(seconds=1), (0.0, 0.0, math.pi/2))
        vb.update_pose(timedelta(seconds=2), (0.0, 0.0, math.pi))
        self.assertTrue(vb.collision())

        # now with angle
        vb = VirtualBumper(timedelta(seconds=2), dist_radius_limit=0.1, angle_limit=math.pi/2)
        vb.update_desired_speed(0.0, math.pi/2)
        self.assertTrue(vb.should_be_moving)
        vb.update_pose(timedelta(seconds=0), (0.0, 0.0, 0.0))
        vb.update_pose(timedelta(seconds=1), (0.0, 0.0, math.pi/2))
        vb.update_pose(timedelta(seconds=2), (0.0, 0.0, math.pi))
        self.assertFalse(vb.collision())

# vim: expandtab sw=4 ts=4



"""
  Draw laser scans
"""
import math
import io
import pathlib
from datetime import timedelta
from collections import defaultdict

import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "1"
import pygame
from pygame.locals import *
import cv2  # for video output
import numpy as np  # faster depth data processing
from importlib import import_module

from osgar.logger import LogIndexedReader, lookup_stream_names
from osgar.lib.serialize import deserialize
from osgar.lib.config import get_class_by_name
from osgar.lib import quaternion
from osgar.lib.depth import depth2danger, DepthParams


WINDOW_SIZE = 1600, 1000  # controlled by --window-size
TAIL_MIN_STEP = 0.1  # in meters
HISTORY_SIZE = 100

MAX_SCAN_LIMIT = 10000  # set by --lidar-limit

g_scale = 30
g_rotation_offset_rad = 0.0  # set by --rotation (deg)
g_lidar_fov_deg = 270  # controlled by --deg (deg)

# depth data for ROBOTIKA_X2_SENSOR_CONFIG_1 (640 x 360)
g_depth_params = DepthParams()

CENTER_AXLE_DISTANCE = 0.348  # K2 distance specific


def scr(x, y):
    global g_scale
    return WINDOW_SIZE[0]//2 + round(g_scale*x), WINDOW_SIZE[1]//2 - round(g_scale*y)


def scan2xy(pose, scan):
    X, Y, heading = pose
    pts = []
    for i, i_dist in enumerate(scan):
        if i_dist == 0 or i_dist >= MAX_SCAN_LIMIT:
            continue
        angle = math.radians(g_lidar_fov_deg * (i / len(scan)) - g_lidar_fov_deg/2) + heading
        dist = i_dist/1000.0
        x, y = dist * math.cos(angle), dist * math.sin(angle)
        pts.append((X + x, Y + y))
    return pts


def filter_pts(grid, pts):
    # filter out duplicity points - keep older
    ret = []
    for x, y in pts:
        k = int(2*x), int(2*y)
        if grid[k] < 10:
            ret.append((x, y))
            grid[k] += 1
    return ret


def draw_scan(foreground, pose, scan, color, joint=None):
    __, __, heading = pose
    X, Y = 0, 0  # front lidar is in the image center

    if joint is not None:
        # correct for joint angle
        dist = CENTER_AXLE_DISTANCE
        heading += math.pi
        dx, dy = dist * math.cos(heading), dist * math.sin(heading)
        X += dx
        Y += dy
        for j in joint:
            heading -= math.radians(j / 100.0)
            dx, dy = dist * math.cos(heading), dist * math.sin(heading)
            X += dx
            Y += dy

    for i, i_dist in enumerate(scan):
        if i_dist == 0 or i_dist >= MAX_SCAN_LIMIT:
            continue
        angle = math.radians(g_lidar_fov_deg * (i / len(scan)) - g_lidar_fov_deg/2) + heading
        dist = i_dist/1000.0
        x, y = dist * math.cos(angle), dist * math.sin(angle)
        pygame.draw.circle(foreground, color, scr(X + x, Y + y), 3)


def draw_scan_up_down(foreground, pose, lidar_up, lidar_down, color):
    __, __, heading = pose

    if lidar_up is not None:
        pygame.draw.circle(foreground, color, scr(0, lidar_up[0]), 3)

    if lidar_down is not None:
        pygame.draw.line(foreground, color, scr(0, 0), scr(0, -lidar_down[0]), 2)
        pygame.draw.circle(foreground, color, scr(0, -lidar_down[0]), 3)


def draw(foreground, pose, scan, poses=[], image=None, bbox=None, callback=None, acc_pts=None):
    foreground.lock()
    color = (0, 255, 0)
    X, Y, heading = pose
    for i, i_dist in enumerate(scan):
        if i_dist == 0 or i_dist >= MAX_SCAN_LIMIT:
            continue
        angle = math.radians(g_lidar_fov_deg * (i / len(scan)) - g_lidar_fov_deg/2) + heading
        dist = i_dist/1000.0
        x, y = dist * math.cos(angle), dist * math.sin(angle)
        pygame.draw.circle(foreground, color, scr(x, y), 3)

    # draw scale
    meter_px = scr(1, 0)[0] - scr(0, 0)[0]
    height = WINDOW_SIZE[1]
    pygame.draw.line(foreground, (255, 0, 0), (20, height - 20), (20 + meter_px, height - 20), 3)
    pygame.draw.line(foreground, (255, 0, 0), (20, height - 10), (20, height - 30), 1)
    pygame.draw.line(foreground, (255, 0, 0), (20 + meter_px, height - 10), (20 + meter_px, height - 30), 1)

    # draw LIDAR pose
    pygame.draw.circle(foreground, (147, 20, 255), scr(0, 0), 5)
    pygame.draw.line(foreground, (147, 20, 255), scr(0, 0), scr(0.5*math.cos(heading), 0.5*math.sin(heading)), 1)

    # draw trace of poses
    for x, y, __ in poses:
        pygame.draw.circle(foreground, (255, 128, 0), scr(x - X, y - Y), 2)

    # draw accumulated points
    if acc_pts is not None:
        for x, y in acc_pts:
            pygame.draw.circle(foreground, (0, 127, 0), scr(x - X, y - Y), 2)

    foreground.unlock()

    if image is not None:
        width, height = image.get_size()
        if width > WINDOW_SIZE[0] or height > WINDOW_SIZE[1]:
            width, height = (512, 384)
        cameraView = pygame.transform.scale(image, (width, height))

        foreground.blit(cameraView, (0, 0))

        for b in bbox:
            assert len(b) > 5, b
            name, x, y, width, height = b[:5]
            color = (0, 255, 0)
            rect = pygame.Rect(x, y, width, height)
            pygame.draw.rect(image, color, rect, 2)

    if callback is not None:
        debug_poly = []
        callback(scan, debug_poly)
        for poly in debug_poly:
            prev = None
            pts = [scr(x, y) for x, y in poly]
            for a, b in zip(pts[:-1], pts[1:]):
                pygame.draw.line(foreground, (200, 200, 0), a, b, 1)


def draw_robot(foreground, pose, joint):

    if joint is not None:
        color = (255, 0, 0)
        __, __, heading = pose

        # first draw line from lidar to joint
        angle = heading + math.pi
        dist = CENTER_AXLE_DISTANCE
        x, y = dist*math.cos(angle), dist*math.sin(angle)
        pygame.draw.line(foreground, color, scr(0, 0), scr(x, y), 3)

        # joint itself
        pygame.draw.circle(foreground, color, scr(x, y), 3)

        # rear part - now K2 specific
        angle -= math.radians(joint[0]/100.0)
        dx, dy = dist*math.cos(angle), dist*math.sin(angle)
        pygame.draw.line(foreground, color, scr(x, y), scr(x + dx, y + dy), 3)

        if len(joint) > 1:
            # K3 specific
            pygame.draw.circle(foreground, color, scr(x + dx, y + dy), 3)
            angle -= math.radians(joint[1]/100.0)
            dx2, dy2 = dist*math.cos(angle), dist*math.sin(angle)
            pygame.draw.line(foreground, color, scr(x + dx, y + dy), scr(x + dx + dx2, y + dy + dy2), 3)


g_depth = None
g_danger_binary_image = False

def get_image(data):
    """Extract JPEG or RGBD depth image"""
    # https://stackoverflow.com/questions/12569452/how-to-identify-numpy-types-in-python
    if isinstance(data, np.ndarray):
        # https://www.learnopencv.com/applycolormap-for-pseudocoloring-in-opencv-c-python/
        if g_danger_binary_image:
            img = np.array(depth2danger(data, g_depth_params) * 255, dtype=np.uint8)
            im_color = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        else:
            img = np.array(np.minimum(255*40, data)/40, dtype=np.uint8)
            im_color = cv2.applyColorMap(img, cv2.COLORMAP_JET)

        # https://stackoverflow.com/questions/19306211/opencv-cv2-image-to-pygame-image
        image = pygame.image.frombuffer(im_color.tostring(), im_color.shape[1::-1], "RGB")
        global g_depth
        g_depth = data
    elif data is not None:
        image = pygame.image.load(io.BytesIO(data), 'JPG').convert()
    else:
        image = None
    return image


def pygame_to_numpy_image(pygame_img):
    view = pygame.surfarray.array3d(pygame_img)
    view = view.transpose([1, 0, 2])
    np_img = cv2.cvtColor(view, cv2.COLOR_RGB2BGR)
    return np_img


def numpy_to_pygame_image(np_image):
    np_img = cv2.cvtColor(np_image, cv2.COLOR_BGR2RGB)
    return pygame.image.frombuffer(np_img.tostring(), np_img.shape[1::-1], "RGB")


class Frame:
    """Just single frame with collected data"""
    def __init__(self):
        self.lidar_up = None
        self.lidar_down = None


class Framer:
    """Creates frames from log entries. Packs together closest scan, pose and camera picture."""
    def __init__(self, filepath, lidar_name=None, lidar2_name=None, pose2d_name=None, pose3d_name=None, camera_name=None,
                 camera2_name=None, bbox_name=None, joint_name=None, keyframes_name=None, title_name=None,
                 lidar_up_name=None, lidar_down_name=None):
        self.log = LogIndexedReader(filepath)
        self.current = 0
        self.frame = Frame()
        self.pose = [0, 0, 0]
        self.pose2d = [0, 0, 0]
        self.pose3d = [[0, 0, 0],[1, 0, 0, 0]]
        self.scan = []
        self.scan2 = []
        self.image = None
        self.image2 = None
        self.bbox = None
        self.joint = None
        self.keyframe = None
        self.title = None
        self.lidar_id, self.pose2d_id, self.pose3d_id, self.camera_id = None, None, None, None
        self.lidar2_id = None
        self.camera2_id = None
        self.bbox_id = None
        self.joint_id = None
        self.keyframes_id = None
        self.title_id = []
        self.lidar_up_id = None
        self.lidar_down_id = None
        names = lookup_stream_names(filepath)
        if lidar_name is not None:
            self.lidar_id = names.index(lidar_name) + 1
        if lidar2_name is not None:
            self.lidar2_id = names.index(lidar2_name) + 1
        if pose2d_name is not None:
            self.pose2d_id = names.index(pose2d_name) + 1
        if pose3d_name is not None:
            self.pose3d_id = names.index(pose3d_name) + 1
        if camera_name is not None:
            self.camera_id = names.index(camera_name) + 1
        if camera2_name is not None:
            self.camera2_id = names.index(camera2_name) + 1
        if bbox_name is not None:
            self.bbox_id = names.index(bbox_name) + 1
        if joint_name is not None:
            self.joint_id = names.index(joint_name) + 1
        if keyframes_name is not None:
            self.keyframes_id = names.index(keyframes_name) + 1
        if title_name is not None:
            title_list = title_name.split(",")
            for ttn in title_list:
                self.title_id.append(names.index(ttn) + 1)
        if lidar_up_name is not None:
            self.lidar_up_id = names.index(lidar_up_name) + 1
        if lidar_down_name is not None:
            self.lidar_down_id = names.index(lidar_down_name) + 1

    def __enter__(self):
        self.log.__enter__()
        return self

    def __exit__(self, *args):
        self.log.__exit__()

    def prev(self):
        return self._step(-1)

    def next(self):
        return self._step(1)

    def seek(self, desired_timestamp):
        print('Seek:', desired_timestamp)
        timestamp, stream_id, data = self.log[self.current]
        print('current', timestamp)
        if timestamp > desired_timestamp:
            return timestamp
        start = self.current
        end = len(self.log) - 1
        timestamp, stream_id, data = self.log[end]
        print('end', timestamp)
        while timestamp < desired_timestamp:
            self.log.grow()
            end = len(self.log) - 1
            timestamp, stream_id, data = self.log[end]
            print('end', timestamp)
        while start + 1 < end:
            mid = (start + end)//2
            timestamp, stream_id, data = self.log[mid]
            if timestamp < desired_timestamp:
                start = mid
            else:
                end = mid
        print('selected', timestamp)
        self.current = start



    def _step(self, direction):
        self.bbox = []
        self.title = []
        if (self.current + direction) >= len(self.log):
            self.log.grow()
        while self.current + direction >= 0 and self.current + direction < len(self.log):
            self.current += direction
            timestamp, stream_id, data = self.log[self.current]
            if stream_id == self.keyframes_id:
                self.keyframe = True
            if stream_id in self.title_id and stream_id != self.bbox_id:
                self.title.append(deserialize(data))
            if stream_id == self.lidar_id:
                self.scan = deserialize(data)
                keyframe = self.keyframe
                self.keyframe = False
                return timestamp, self.frame, self.pose, self.pose3d, self.scan, self.scan2, self.image, self.image2, self.bbox, self.joint, keyframe, self.title, False
            if stream_id == self.lidar2_id:
                self.scan2 = deserialize(data)
            if stream_id == self.lidar_up_id:
                self.frame.lidar_up = deserialize(data)
            if stream_id == self.lidar_down_id:
                self.frame.lidar_down = deserialize(data)
            elif stream_id == self.camera_id:
                self.image = get_image(deserialize(data))
                # bounding boxes associated with an image are stored after the image in the log
                # therefore, we need to continue reading the log past the image in order to gathering its bounding box data
                current = self.current
                while current + direction >= 0 and current + direction < len(self.log):
                    current += direction
                    _, new_stream_id, new_data = self.log[current]
                    if new_stream_id == self.bbox_id:
                        self.bbox.append(deserialize(new_data))
                    if new_stream_id in self.title_id:
                        self.title.append(deserialize(new_data))
                    if new_stream_id == self.camera_id:
                        break
                if self.lidar_id is None:
                    keyframe = self.keyframe
                    self.keyframe = False
                    return timestamp, self.frame, self.pose, self.pose3d, self.scan, self.scan2, self.image, self.image2, self.bbox, self.joint, keyframe, self.title, False
            elif stream_id == self.camera2_id:
                self.image2 = get_image(deserialize(data))
            elif stream_id == self.joint_id:
                self.joint = deserialize(data)
            elif stream_id == self.pose3d_id:
                pose3d, orientation = deserialize(data)
                assert len(pose3d) == 3
                assert len(orientation) == 4
                self.pose = [pose3d[0], pose3d[1], quaternion.heading(orientation)]
                self.pose3d = [pose3d, orientation]
            elif stream_id == self.pose2d_id:
                arr = deserialize(data)
                assert len(arr) == 3
                self.pose = (arr[0]/1000.0, arr[1]/1000.0, math.radians(arr[2]/100.0))
                x, y, heading = self.pose
                self.pose = (x * math.cos(g_rotation_offset_rad) - y * math.sin(g_rotation_offset_rad),
                             x * math.sin(g_rotation_offset_rad) + y * math.cos(g_rotation_offset_rad),
                             heading + g_rotation_offset_rad)
                if self.lidar_id is None and self.camera_id is None:
                    keyframe = self.keyframe
                    self.keyframe = False
                    return timestamp, self.frame, self.pose, self.pose3d, self.scan, self.scan2, self.image, self.image2, self.bbox, self.joint, keyframe, self.title, False
        return timedelta(), self.frame, self.pose, self.pose3d, self.scan, self.scan2, self.image, self.image2, self.bbox, self.joint, self.keyframe, self.title, True


def lidarview(gen, caption_filename, callback=False, callback_img=False, out_video=None, jump=None):
    global g_scale, WINDOW_SIZE
    last_timestamp = None
    last_image = None

    if out_video is not None:
        width, height = WINDOW_SIZE
        fps = 10
        writer = cv2.VideoWriter(out_video,
                                 cv2.VideoWriter_fourcc('F', 'M', 'P', '4'),
                                 fps,
                                 (width, height))

    pygame.display.init()
    screen = pygame.display.set_mode(WINDOW_SIZE, pygame.RESIZABLE)

    # create backgroud
    background = pygame.Surface(screen.get_size())
#    background.set_colorkey((0,0,0))

    # create foreground
    foreground = pygame.Surface(screen.get_size())
#    foreground.set_colorkey((0,0,0))

    # display everything
    screen.blit(background, (0, 0))
    screen.blit(foreground, (0, 0))
    pygame.display.flip()

    pygame.key.set_repeat(200, 20)

    paused = False
    camera_on = True
    use_image2 = False
    map_on = False
    poses = []
    acc_pts = []
    grid = defaultdict(int)
    skip_frames = 0
    frames_step = 0
    save_counter = 0
    was_resized = False

    history = gen
    max_timestamp = None
    wait_for_keyframe = False

    if jump is not None:
        gen.seek(timedelta(seconds=jump))

    while True:
        timestamp, frame, pose, pose3d, scan, scan2, image, image2, bbox, joint, keyframe, title, eof = history.next()

        if max_timestamp is None or max_timestamp < timestamp:
            # build map only for new data
            max_timestamp = timestamp

            # remove potential duplicity of poses (i.e. when not moving)
            if len(poses) == 0 or math.hypot(poses[-1][0] - pose[0], poses[-1][1] - pose[1]) >= TAIL_MIN_STEP:
                poses.append(pose)

            xy_scan = scan2xy(pose, scan)
            xy_scan_filtred = filter_pts(grid, xy_scan)
            acc_pts.extend(xy_scan_filtred)

        if wait_for_keyframe and not keyframe and not eof:
            paused = True
            caption = caption_filename + ": %s" % timestamp
            caption += ' searching fwd ...'
            pygame.display.set_caption(caption)
            event = pygame.event.poll()
            if event.type != KEYDOWN:
                continue
        wait_for_keyframe = False

        if skip_frames > 0 and not eof:
            skip_frames -= 1
            continue
        skip_frames = frames_step

        while True:
            caption = caption_filename + ": %s" % timestamp
            for t in title:
                caption += ' (' + str(t) + ')'
            if paused:
                caption += ' (PAUSED)'
            if frames_step > 0:
                caption += ' step=' + str(frames_step)
            if eof:
                caption += ' [END]'
            if keyframe:
                caption += ' [KEYFRAME]'
            pygame.display.set_caption(caption)

            foreground.fill((0, 0, 0))
            img = image2 if use_image2 else image
            if callback_img:
                if last_timestamp != timestamp:
                    img = numpy_to_pygame_image(callback_img(pygame_to_numpy_image(img)))
                    last_timestamp = timestamp
                    last_image = img
                else:
                    img = last_image
            draw_robot(foreground, pose, joint)
            draw_scan(foreground, pose, scan2, color=(128, 128, 0), joint=joint)
            draw_scan_up_down(foreground, pose, frame.lidar_up, frame.lidar_down, color=(255, 0, 0))
            draw(foreground, pose, scan, poses=poses,
                 image=img if camera_on else None, bbox=bbox,
                 acc_pts=acc_pts if map_on else None,
                 callback=callback)
            screen.blit(background, (0, 0))
            screen.blit(foreground, (0, 0))
            pygame.display.flip()

            if out_video is not None:
                # https://stackoverflow.com/questions/53101698/how-to-convert-a-pygame-image-to-open-cv-image/53108946#53108946
                #  create a copy of the surface
                view = pygame.surfarray.array3d(screen)
                #  convert from (width, height, channel) to (height, width, channel)
                view = view.transpose([1, 0, 2])
                #  convert from rgb to bgr
                img_bgr = cv2.cvtColor(view, cv2.COLOR_RGB2BGR)
                writer.write(img_bgr)

            if paused or eof:
                event = pygame.event.wait()
            else:
                event = pygame.event.poll()
            if event.type == QUIT:
                return
            if event.type == pygame.VIDEORESIZE:
                was_resized = True
                updated_size = event.size
            if event.type is pygame.ACTIVEEVENT and was_resized:
                was_resized = False
                screen = pygame.display.set_mode(updated_size, pygame.RESIZABLE)
                WINDOW_SIZE = updated_size
                background = pygame.Surface(screen.get_size())
                foreground = pygame.Surface(screen.get_size())
            if event.type == KEYDOWN:
                if event.key in [K_ESCAPE, K_q]:
                    return
                if event.key == K_SPACE:
                    paused = not paused
                if event.key in [K_PLUS, K_KP_PLUS, K_EQUALS]:
                    # K_EQUALS is for convenience: on notebook keyboard is '+' as Shift+'='
                    g_scale *= 2.0
                if event.key in [K_MINUS, K_KP_MINUS]:
                    g_scale /= 2.0
                if event.key == K_c:
                    if pygame.key.get_mods() & pygame.KMOD_LSHIFT:
                        use_image2 = not use_image2
                    else:
                        camera_on = not camera_on
                if event.key == K_m:
                    map_on = not map_on
                if event.key == K_0:
                    frames_step = 0
                if event.key == K_1:
                    frames_step = 10
                if event.key == K_2:
                    frames_step = 20
                if event.key == K_3:
                    frames_step = 30
                if event.key == K_4:
                    frames_step = 40
                if event.key == K_9:
                    frames_step = 90
                if event.key == K_s:
                    save_image = image2 if use_image2 else image
                    pygame.image.save(save_image, "saveX-{:04}.jpg".format(save_counter))
                    save_counter += 1
                if event.key == K_b:  # swap binary danger image on/off
                    global g_danger_binary_image
                    g_danger_binary_image = not g_danger_binary_image
                    history.prev()
                if event.key == K_d:  # dump scan
                    print(scan)
                    if g_depth is not None:
                        np.savez_compressed('depth.npz', depth=g_depth, pose3d=pose3d, img=pygame_to_numpy_image(image), scan=scan)
                if event.key == K_p:  # print position
                    print(pose)
                    x, y, heading = pose
                    if math.hypot(x, y) > 0.1:
                        print('rotation (deg) =', math.degrees(heading), 'dist =', math.hypot(x, y))
                    else:
                        print('rotation not available')
                if event.key == K_n:  # next keyframe
                    aborted = False
                    if pygame.key.get_mods() & pygame.KMOD_LSHIFT:  # previous keyframe
                        if keyframe:
                            history.prev()  # search for the previous one
                        data = history.prev()
                        while not data[-2] and not data[-1]:
                            # EOF is at the beginning as well as at the end
                            data = history.prev()
                            caption = caption_filename + ": %s" % data[0]
                            caption += ' searching bwd ...'
                            pygame.display.set_caption(caption)
                            tmp_event = pygame.event.poll()
                            if tmp_event.type == KEYDOWN:
                                aborted = True
                                timestamp = data[0]
                                break
                        history.prev()
                    if aborted:
                        paused = True
                    else:
                        wait_for_keyframe = True
                        paused = False  # let it search for next keyframe

                if event.key == K_RIGHT:
                    break
                if event.key == K_LEFT:
                    paused = True
                    history.prev()
                    history.prev()
                    break
            if event.type == pygame.NOEVENT and not paused and not eof:
                break
    if out_video is not None:
        writer.release()


def main(args_in=None, startswith=None):
    import argparse
    import os.path
    global g_rotation_offset_rad, g_lidar_fov_deg, MAX_SCAN_LIMIT, WINDOW_SIZE

    parser = argparse.ArgumentParser(description='View lidar scans')
    parser.add_argument('logfile', help='recorded log file')

    parser.add_argument('--lidar', help='stream ID')
    parser.add_argument('--lidar2', help='stream ID of second lidar (back or slope)')
    parser.add_argument('--lidar-limit', help='display scan limit in millimeters',
                        type=int, default=MAX_SCAN_LIMIT)
    parser.add_argument('--lidar-up', help='drone point lidar up stream ID')
    parser.add_argument('--lidar-down', help='drone point lidar down stream ID')
    pose = parser.add_mutually_exclusive_group()
    pose.add_argument('--pose2d', help='stream ID for pose2d messages')
    pose.add_argument('--pose3d', help='stream ID for pose3d messages')

    parser.add_argument('--camera', help='stream ID for JPEG images')
    parser.add_argument('--camera2', help='stream ID for 2nd JPEG images')
    parser.add_argument('--bbox', help='stream ID for detection bounding box')

    parser.add_argument('--joint', help='stream ID joint angle for articulated robots (Kloubak)')

    parser.add_argument('--keyframes', help='stream ID typically for artifacts detection')
    parser.add_argument('--title', help='stream ID of data to be displayed in title')

    parser.add_argument('--window-size', help='set window size in pixels', type=int, nargs=2)

    parser.add_argument('--callback', help='callback function for lidar scans')

    parser.add_argument('--callback-img', help='callback function for image, eg.: subt.tf_detector:TfDetector')

    parser.add_argument('--rotate', help='rotate poses by angle in degrees, offset',
                        type=float, default=0.0)

    parser.add_argument('--deg', help='lidar field of view in degrees',
                        type=float, default=270)

    parser.add_argument('--jump', '-j', help='jump to given time in seconds', type=int)

    parser.add_argument('--create-video', help='filename of output video')

    args = parser.parse_args(args_in)

    p = pathlib.Path(args.logfile)
    if p.is_dir():
        if startswith is not None:
            g = iter(child for child in p.iterdir() if child.name.startswith(startswith))
        else:
            g = p.iterdir()
        args.logfile = max(g, key=lambda a: a.stat().st_mtime)
        print(args.logfile)

    if not any([args.lidar, args.pose2d, args.pose3d, args.camera]):
        print("Available streams:")
        for stream in lookup_stream_names(args.logfile):
            print("  ", stream)
        return

    callback = None
    if args.callback is not None:
        callback = get_class_by_name(args.callback)
    callback_img = None
    if args.callback_img:
        name = args.callback_img
        assert ':' in name, name  # import path and class name expected
        s = name.split(':')
        assert len(s) == 2, s  # package and class name
        module_name, class_name = s
        m = import_module(module_name)
        callback_img = getattr(m, class_name)
        callback_img = callback_img().run_on_image

    if args.lidar_limit is not None:
        MAX_SCAN_LIMIT = args.lidar_limit
    if args.window_size is not None:
        WINDOW_SIZE = args.window_size

    filename = os.path.basename(args.logfile)
    g_rotation_offset_rad = math.radians(args.rotate)
    g_lidar_fov_deg = args.deg
    with Framer(args.logfile, lidar_name=args.lidar, lidar2_name=args.lidar2, pose2d_name=args.pose2d, pose3d_name=args.pose3d,
                camera_name=args.camera, camera2_name=args.camera2, bbox_name=args.bbox, joint_name=args.joint,
                keyframes_name=args.keyframes, title_name=args.title, lidar_up_name=args.lidar_up, lidar_down_name=args.lidar_down) as framer:
        lidarview(framer, caption_filename=filename, callback=callback, callback_img=callback_img, out_video=args.create_video, jump=args.jump)

if __name__ == "__main__":
    main()

# vim: expandtab sw=4 ts=4



import unittest
import math

from subt.local_planner import LocalPlanner, LocalPlannerRef, LocalPlannerOpt, LocalPlannerNumpy


class SubTChallengeTest(unittest.TestCase):

    def test_usage(self):
        planner = LocalPlanner()
        scan = [0] * 270  # no obstacles
        planner.update(scan)
        self.assertEqual(planner.recommend(0), (1.0, 0.0))

        scan[135] = 1000
        scan[100] = 1000  # the direction from right to left
        planner.update(scan)
        good, direction = planner.recommend(0)
        self.assertAlmostEqual(good, 0.47525253623471064)
        self.assertAlmostEqual(direction, 1.3089969389957472)

        # extra obstacles should not influence the decision
        for i in range(135):
            scan[i] = 1000
        planner.update(scan)
        good, direction = planner.recommend(0)
        self.assertAlmostEqual(good, 0.47525253623471064)
        self.assertAlmostEqual(direction, 1.3089969389957472)

    def test_corner(self):
        planner = LocalPlannerOpt()
        # obstacle defined as line "x + y - 1 = 0", corner defined via abs()
        # ray defined as t*cos(angle), t*sin(angle)
        # solution for t = 1/(cos(angle) + sin(angle))
        angles = [abs(math.radians(i-135)) for i in range(271)]
        scan = [int(1400 / (math.cos(a) + math.sin(a))) for a in angles]
        planner.update(scan)
        self.assertEqual(planner.recommend(0), (0.18451952399298924, 2.0420352248333655))

        # now test also old planner
        planner_ref = LocalPlannerRef()
        planner_ref.update(scan)
        self.assertEqual(planner.recommend(0), planner_ref.recommend(0))

        # but why numpy is correct?
        planner = LocalPlannerNumpy()
        planner.update(scan)
        self.assertEqual(planner.recommend(0), (0.18451952399298924, -2.0420352248333655))

# vim: expandtab sw=4 ts=4




#!/usr/bin/python
"""
  Convert logfile to AVI video
"""
import pathlib
from datetime import timedelta

try:
    import cv2
except ImportError:
    print('\nERROR: Please install OpenCV\n    pip install opencv-python\n')

try:
    import numpy as np
except ImportError:
    print('\nERROR: Please install numpy\n    pip install numpy\n')

from osgar.logger import LogReader, lookup_stream_id
from osgar.lib.serialize import deserialize


def create_video(logfile, stream, outfile, add_time=False,
                 start_time_sec=0, end_time_sec=None, fps=25,
                 flip=False, camera2=None):
    if outfile is None:
        outfile = str(pathlib.Path(logfile).with_suffix(".mp4"))
    assert outfile.endswith(".mp4"), outfile
    only_stream = lookup_stream_id(logfile, stream)
    if camera2 is not None:
        dual_cam_id = lookup_stream_id(logfile, camera2)
        only_stream = [only_stream, dual_cam_id]
        img1, img2 = None, None
    else:
        dual_cam_id = None
    with LogReader(logfile, only_stream_id=only_stream) as log:
        writer = None
        for timestamp, stream_id, data in log:
            buf = deserialize(data)
            img = cv2.imdecode(np.fromstring(buf, dtype=np.uint8), 1)
            if flip:
                img = cv2.flip(img, 0)
            if dual_cam_id is not None:
                if stream_id == dual_cam_id:
                    img2 = img
                else:
                    img1 = img
                # make sure that both images are defined before you create writer
                if img1 is None or img2 is None:
                    continue
                img = np.concatenate((img1, img2), axis=1)
            if writer is None:
                height, width = img.shape[:2]
                writer = cv2.VideoWriter(outfile,
                                         cv2.VideoWriter_fourcc(*"mp4v"),
                                         fps,
                                         (width, height))
            if add_time:
                if (width, height) == (640, 480):
                    x, y = 350, 50
                    thickness = 3
                    size = 3.0
                else:
                    x, y = 800, 100
                    thickness = 5
                    size = 5.0
                # clip microseconds to miliseconds
                s = str(timestamp-timedelta(seconds=start_time_sec))[:-3]
                cv2.putText(img, s, (x, y), cv2.FONT_HERSHEY_PLAIN, 
                            size, (255, 255, 255), thickness=thickness)

            if timestamp.total_seconds() > start_time_sec:
                writer.write(img)
            if end_time_sec is not None and timestamp.total_seconds() > end_time_sec:
                break
        writer.release()


def main():
    import argparse

    parser = argparse.ArgumentParser(description='Convert logfile to AVI video')
    parser.add_argument('logfile', help='recorded log file')
    parser.add_argument('--stream', help='stream ID or name', default='camera.raw')
    parser.add_argument('--camera2', help='optional 2nd camera stream ID or name')
    parser.add_argument('--out', '-o', help='output AVI file', default=None)
    parser.add_argument('--display-time', '-t', help='add timestamp info', action='store_true')
    parser.add_argument('--start-time-sec', '-s', help='start video at later time (sec)',
                        type=float, default=0.0)
    parser.add_argument('--end-time-sec', '-e', help='cut video at given time (sec)',
                        type=float, default=None)
    parser.add_argument('--fps', help='frames per second', type=int, default=25)
    parser.add_argument('--flip', help='horizontal flip', action='store_true')
    args = parser.parse_args()

    create_video(args.logfile, args.stream, args.out, add_time=args.display_time,
                 start_time_sec=args.start_time_sec, end_time_sec=args.end_time_sec, fps=args.fps,
                 flip=args.flip, camera2=args.camera2)


if __name__ == "__main__":
    main()

# vim: expandtab sw=4 ts=4 




import unittest
from unittest.mock import MagicMock, patch
from time import sleep

from subt.wifisignal import WifiSignal


class WifiSignalTest(unittest.TestCase):

    def test_usage(self):
        with patch('subt.wifisignal.wifi_scan') as p:
            bus = MagicMock()
            ws = WifiSignal(config={}, bus=bus)
            bus.is_alive = MagicMock(return_value=True)
            ws.start()
            sleep(0.1)
            bus.is_alive = MagicMock(return_value=False)
            ws.join()

# vim: expandtab sw=4 ts=4




import math
import numpy as np
import cProfile


g_pr = cProfile.Profile()
g_count = 0


def normalize_angle(angle):
    return (angle + math.pi) % (2 * math.pi) - math.pi


class LocalPlannerRef:
    def __init__(self, scan_start=math.radians(-135), scan_end=math.radians(135), direction_adherence=math.radians(90),
                 max_obstacle_distance=1.5, obstacle_influence=1.2, scan_subsample=1, max_considered_obstacles=None):
        self.last_scan = None
        self.scan_start = scan_start
        self.scan_end = scan_end
        self.direction_adherence = direction_adherence
        self.max_obstacle_distance = max_obstacle_distance
        self.obstacle_influence = obstacle_influence
        self.scan_subsample = scan_subsample

    def update(self, scan):
        self.last_scan = scan[::self.scan_subsample] if self.scan_subsample > 1 else scan

    def recommend(self, desired_dir):
        if self.last_scan is None:
            return 1.0, desired_dir

        obstacles = []
        for (i, measurement) in enumerate(self.last_scan):
            if measurement == 0:
                continue
            if measurement * 1e-3 > self.max_obstacle_distance:
                continue
            measurement_angle = self.scan_start + (self.scan_end - self.scan_start) * i / float(len(self.last_scan) - 1)
            measurement_vector = math.cos(measurement_angle), math.sin(measurement_angle)

            # Converting from millimeters to meters.
            obstacle_xy = [mv * (measurement * 1e-3) for mv in measurement_vector]

            obstacles.append(obstacle_xy)

        if not obstacles:
            return 1.0, normalize_angle(desired_dir)

        # Best direction points roughly in desired_dir and does not come too close to any obstacle.
        def is_desired(direction):
            direction_delta = normalize_angle(direction - desired_dir)
            return math.exp(-(direction_delta / self.direction_adherence)**2)  # Fuzzy equality

        def is_risky(direction):
            direction_vector = math.cos(direction), math.sin(direction)
            riskiness = 0.
            for obstacle_xy in obstacles:
                # Obstacles behind the robot from the perspective of the desired direction do not matter.
                # The formula below computes cos(angle between the two vectors) * their_norms. Norms are positive, so a negative result implies abs(angle) > 90deg.
                if sum(d * o for (d, o) in zip(direction_vector, obstacle_xy)) < 0:
                    continue

                # Distance between the obstacle and line defined by direction.
                # https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
                # Norm of direction_vector is 1.0, so we do not need to divide by it.
                off_track_distance = abs(direction_vector[1] * obstacle_xy[0] - direction_vector[0] * obstacle_xy[1])

                r = math.exp(-(off_track_distance / self.obstacle_influence)**2)
                if r > riskiness: # max as fuzzy OR.
                    riskiness = r

            return riskiness

        def is_safe(direction):
            return 1.0 - is_risky(direction)  # Fuzzy negation.

        def is_good(direction):
            return min(is_safe(direction), is_desired(direction))  # Fuzzy AND.

        return max((is_good(math.radians(direction)), math.radians(direction)) for direction in range(-180, 180, 3))


class LocalPlannerOpt:
    """
    Optimized version of Local Planner.
    1) ignore all directions into obstacles - there is always a way out (lidar is 270 degrees so it can go freely back)
       and way into the obstacle will have high penalization cost.
    2) leave complex evaluation as the last step for already best direction
    """
    def __init__(self, scan_start=math.radians(-135), scan_end=math.radians(135), direction_adherence=math.radians(90),
                 max_obstacle_distance=1.5, obstacle_influence=1.2, scan_subsample=1, max_considered_obstacles=None):
        self.last_scan = None
        self.scan_start = scan_start
        self.scan_end = scan_end
        self.direction_adherence = direction_adherence
        self.max_obstacle_distance = max_obstacle_distance
        self.obstacle_influence = obstacle_influence
        self.scan_subsample = scan_subsample

    def update(self, scan):
        self.last_scan = scan[::self.scan_subsample] if self.scan_subsample > 1 else scan

    def recommend(self, desired_dir):
        if self.last_scan is None:
            return 1.0, desired_dir

        # valid candidate angles for consideration
        valid = [True] * 361  # indexed -180 .. 180

        obstacles = []
        for (i, measurement) in enumerate(self.last_scan):
            if measurement == 0:
                continue
            if measurement * 1e-3 > self.max_obstacle_distance:
                continue
            measurement_angle = self.scan_start + (self.scan_end - self.scan_start) * i / float(len(self.last_scan) - 1)
            measurement_vector = math.cos(measurement_angle), math.sin(measurement_angle)

            # Converting from millimeters to meters.
            obstacle_xy = [mv * measurement * 1e-3 for mv in measurement_vector]

            obstacles.append(obstacle_xy)
            valid[int(math.degrees(normalize_angle(measurement_angle))) + 180] = False

#        print('obstacles', len(obstacles))
        if not obstacles:
            return 1.0, normalize_angle(desired_dir)

        # Best direction points roughly in desired_dir and does not come too close to any obstacle.
        def is_desired(direction):
            direction_delta = normalize_angle(direction - desired_dir)
            return math.exp(-(direction_delta / self.direction_adherence)**2)  # Fuzzy equality

        def is_risky(direction):
            direction_vector = math.cos(direction), math.sin(direction)
            min_off_track_distance = 1000.
            for obstacle_xy in obstacles:
                # Obstacles behind the robot from the perspective of the desired direction do not matter.
                # The formula below computes cos(angle between the two vectors) * their_norms. Norms are positive, so a negative result implies abs(angle) > 90deg.
                if sum(d * o for (d, o) in zip(direction_vector, obstacle_xy)) < 0:
                    continue

                # Distance between the obstacle and line defined by direction.
                # https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
                # Norm of direction_vector is 1.0, so we do not need to divide by it.
                off_track_distance = abs(direction_vector[1] * obstacle_xy[0] - direction_vector[0] * obstacle_xy[1])

                if off_track_distance < min_off_track_distance: # max as fuzzy OR.
                    min_off_track_distance = off_track_distance

            riskiness = math.exp(-(min_off_track_distance / self.obstacle_influence)**2)
            return riskiness

        def is_safe(direction):
            return 1.0 - is_risky(direction)  # Fuzzy negation.

        def is_good(direction):
            return min(is_safe(direction), is_desired(direction))  # Fuzzy AND.

        return max((is_good(math.radians(direction)), math.radians(direction)) for direction in range(-180, 180, 3) if valid[direction + 180])

class LocalPlannerNumpy:
    def __init__(self, scan_start=math.radians(-135), scan_end=math.radians(135), direction_adherence=math.radians(90),
                 max_obstacle_distance=1.5, obstacle_influence=1.2, scan_subsample=1, max_considered_obstacles=None):
        self.last_scan = None
        self.scan_start = scan_start
        self.scan_end = scan_end
        self.direction_adherence = direction_adherence
        self.max_obstacle_distance = max_obstacle_distance
        self.obstacle_influence = obstacle_influence

        self.considered_directions = np.radians(
                np.linspace(
                    -180, +180, 360//3, endpoint=False)).reshape((-1, 1))
        self.considered_directions_cos = np.cos(self.considered_directions)
        self.considered_directions_sin = np.sin(self.considered_directions)
        self.scan_subsample = scan_subsample

        # To be filled in later, once we know how many directions per scan we
        # receive.
        self.angles = None
        self.angles_cos = None
        self.angles_sin = None
        self.scan = None
        self.max_considered_obstacles = max_considered_obstacles

    def update(self, scan):
        self.scan = np.asarray(scan[::self.scan_subsample] if self.scan_subsample > 1 else scan) * 1e-3

        if self.angles is None or len(self.scan) != self.angles.shape[0]:
            n = len(self.scan)
            delta = (self.scan_end - self.scan_start) / (n - 1)
            self.angles = np.linspace(
                    self.scan_start,
                    self.scan_end,
                    n).reshape((1, -1))
            self.angles_cos = np.cos(self.angles)
            self.angles_sin = np.sin(self.angles)

    def recommend(self, desired_dir):
        if self.scan is None:
            return 0.0, desired_dir
        NO_MEASUREMENT = 0
        is_valid = np.logical_and(
                self.scan != NO_MEASUREMENT,
                self.scan <= self.max_obstacle_distance)
        valid_scan = self.scan[is_valid]
        is_valid = is_valid.reshape((1, -1))
        acoss = self.angles_cos[is_valid]
        asins = self.angles_sin[is_valid]
        x = acoss * valid_scan
        y = asins * valid_scan

        if x.shape[0] == 0:
            return 1.0, normalize_angle(desired_dir)

        if self.max_considered_obstacles is not None and x.shape[0] > self.max_considered_obstacles:
            nth = math.ceil(x.shape[0] / self.max_considered_obstacles)
            x = x[::nth]
            y = y[::nth]

        num_considered_directions = self.considered_directions.shape[0]
        xs = np.broadcast_to(x, (num_considered_directions, x.shape[0]))
        ys = np.broadcast_to(y, (num_considered_directions, y.shape[0]))
        dist_from_line = np.abs(xs * self.considered_directions_sin -
                                ys * self.considered_directions_cos)
        # Obstacles behind the robot from the perspective of the desired direction do not matter.
        # The formula below computes cos(angle between the two vectors) * their_norms. Norms are positive, so a negative result implies abs(angle) > 90deg.
        is_in_front = (xs * self.considered_directions_cos +
                       ys * self.considered_directions_sin) >= 0
        # Only valid points that can block the movement count.
        dist_from_line = np.where(is_in_front, dist_from_line, float('inf'))
        # In each direction, we only care about the nearest obstacle.
        min_dist = np.min(dist_from_line, axis=1)
        risky = np.exp(-(min_dist / self.obstacle_influence)**2)
        safe = 1.0 - risky  # Fuzzy negation.

        direction_delta = normalize_angle(self.considered_directions - desired_dir)
        desired = np.exp(-(direction_delta/self.direction_adherence)**2).reshape((-1,))  # Fuzzy equality

        good = np.minimum(safe, desired)  # Fuzzy and.
        best = np.argmax(good)
        return good[best], self.considered_directions[best][0]


class LocalPlanner:
    def __init__(self, *args, **kwargs):
        self.opt = LocalPlannerOpt(*args, **kwargs)
        self.ref = LocalPlannerRef(*args, **kwargs)
        self.nump = LocalPlannerNumpy(*args, **kwargs)

    def update(self, scan):
        self.opt.update(scan)
        self.ref.update(scan)
        self.nump.update(scan)

    def recommend(self, desired_dir):
#        global g_count
#        g_count += 1
#        g_pr.enable()
        ret = self.nump.recommend(desired_dir)
#        ret = self.opt.recommend(desired_dir)
#        ref = self.ref.recommend(desired_dir)
#        g_pr.disable()
#        if g_count % 100 == 0:
#            g_pr.print_stats()
#        assert ret == ref, (ret, ref)
        return ret

# vim: expandtab sw=4 ts=4



import datetime
import unittest
from unittest.mock import MagicMock, patch, call

from osgar.bus import Bus
from subt.teambase import Teambase


class TeambaseTest(unittest.TestCase):

    def test_finish_time(self):
        tb = Teambase(bus=MagicMock(), config={'robot_name':'T42'})
        self.assertEqual(tb.finish_time, 42)

    def test_on_robot_xyz(self):
        tb = Teambase(bus=MagicMock(), config={'robot_name':'T300'})
        self.assertEqual(tb.robot_positions, {})
        data = ['A0F150L', [11896, 7018, -11886]]
        tb.on_robot_xyz(data)
        self.assertEqual(tb.robot_positions, {'A0F150L': [11896, 7018, -11886]})

# vim: expandtab sw=4 ts=4



"""
  Driver for CO2 sensor via USB
  Intelligent Infrared CO2 Module
    (Model: MH-Z19)
"""

import struct

from osgar.node import Node
from osgar.bus import BusShutdownException


class WinsenCO2(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('raw', 'co2')
        self.sleep_time = config.get('sleep')
        self._buf = b''
        self.errors = 0

    def create_packet(self):
        # request CO2 readings
        return b"\xff\x01\x86\x00\x00\x00\x00\x00\x79"

    def get_packet(self):
        """extract packet from internal buffer (if available otherwise return None"""
        data = self._buf
        if 0xFF not in data:
            self._buf = b''
            return None
        i = data.index(0xFF)
        data = data[i:]  # cut to the beginning of the packet
        size = 9
        if len(data) < size:
            self._buf = data
            return None
        ret, self._buf = data[:size], data[size:]
        checksum = sum(ret[1:]) & 0xFF
        if checksum != 0:
            self.errors += 1
            return None
        return ret

    def parse_CO2_packet(self, data):
        """
        Parse CO2 value
        """
        return struct.unpack_from('>H', data, 2)[0]

    def run(self):
        try:
            while True:
                dt, channel, data = self.listen()
                self.time = dt
                if channel == 'trigger':
                    self.publish('raw', self.create_packet())
                elif channel == 'raw':
                    self._buf += data
                    packet = self.get_packet()
                    if packet is not None:
                        value_CO2 = self.parse_CO2_packet(packet)
                        if value_CO2 is None:
                            print(packet)
                        else:
                            self.publish('co2', value_CO2)

        except BusShutdownException:
            pass

        assert self.errors == 0, self.errors  # checksum

# vim: expandtab sw=4 ts=4



"""
  Quaternion - defined as (x, y, z, w) to be compatible with ROS and RealSense

  ROS:
    http://wiki.ros.org/tf2/Tutorials/Quaternions

ROS uses quaternions to track and apply rotations. A quaternion has 4
components (x,y,z,w). That's right, 'w' is last (but beware: some libraries
like Eigen put w as the first number!). The commonly-used unit quaternion that
yields no rotation about the x/y/z axes is (0,0,0,1)

https://www.andre-gaschler.com/rotationconverter/
https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles
"""
import math


def multiply(quaternion1, quaternion0):
    x0, y0, z0, w0 = quaternion0
    x1, y1, z1, w1 = quaternion1
    return [ x1*w0 + y1*z0 - z1*y0 + w1*x0,
            -x1*z0 + y1*w0 + z1*x0 + w1*y0,
             x1*y0 - y1*x0 + z1*w0 + w1*z0,
            -x1*x0 - y1*y0 - z1*z0 + w1*w0]

def conjugate(quaternion):
    x, y, z, w = quaternion
    return [-x, -y, -z, w]

def identity():
    return [0, 0, 0, 1]

def rotate_vector(vector, quaternion):
    # https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation
    qvector = vector + [0]
    con = conjugate(quaternion)
    part1 = multiply(quaternion, qvector)
    return multiply(part1, con)[:-1]


def normalize(quaternion):
    x0, y0, z0, w0 = quaternion
    sqr_size = x0*x0 + y0*y0 + z0*z0 + w0*w0
    if abs(sqr_size - 1.0) > 0.00001:
        k = math.sqrt(sqr_size)
        x0, y0, z0, w0 = x0/k, y0/k, z0/k, w0/k
    return [x0, y0, z0, w0]

def euler_to_quaternion(yaw, pitch, roll):
    qx = math.sin(roll/2) * math.cos(pitch/2) * math.cos(yaw/2) - math.cos(roll/2) * math.sin(pitch/2) * math.sin(yaw/2)
    qy = math.cos(roll/2) * math.sin(pitch/2) * math.cos(yaw/2) + math.sin(roll/2) * math.cos(pitch/2) * math.sin(yaw/2)
    qz = math.cos(roll/2) * math.cos(pitch/2) * math.sin(yaw/2) - math.sin(roll/2) * math.sin(pitch/2) * math.cos(yaw/2)
    qw = math.cos(roll/2) * math.cos(pitch/2) * math.cos(yaw/2) + math.sin(roll/2) * math.sin(pitch/2) * math.sin(yaw/2)
    return [qx, qy, qz, qw]

def euler_zyx(quaternion):
    x0, y0, z0, w0 = normalize(quaternion)
    ax =  math.atan2(2*(w0*x0+y0*z0), 1-2*(x0*x0+y0*y0))
    ay =  math.asin(2*(w0*y0-z0*x0))
    az =  math.atan2(2*(w0*z0+x0*y0), 1-2*(y0*y0+z0*z0))
    return [az, ay, ax]

def heading(quaternion):
    x0, y0, z0, w0 = quaternion
    az =  math.atan2(2*(w0*z0+x0*y0), 1-2*(y0*y0+z0*z0))
    return az

def from_axis_angle(axis, angle):
    ax, ay, az = axis
    qx = ax * math.sin(angle / 2)
    qy = ay * math.sin(angle / 2)
    qz = az * math.sin(angle / 2)
    qw = math.cos(angle / 2)
    return [qx, qy, qz, qw]


def rotation_matrix(quaternion):
    qx, qy, qz, qw = quaternion
    r1 = [1 - 2*qy**2 - 2*qz**2,  2*qx*qy - 2*qz*qw,      2*qx*qz + 2*qy*qw]
    r2 = [2*qx*qy + 2*qz*qw,      1 - 2*qx**2 - 2*qz**2,  2*qy*qz - 2*qx*qw]
    r3 = [2*qx*qz - 2*qy*qw,      2*qy*qz + 2*qx*qw,      1 - 2*qx**2 - 2*qy**2]
    return [r1, r2, r3]


def from_rotation_matrix(rotation_matrix):
    # http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm
    # https://en.wikipedia.org/wiki/Rotation_matrix#Quaternion
    m00, m01, m02 = rotation_matrix[0]
    m10, m11, m12 = rotation_matrix[1]
    m20, m21, m22 = rotation_matrix[2]
    qw = math.sqrt(max(0, 1 + m00 + m11 + m22)) / 2
    qx = math.sqrt(max(0, 1 + m00 - m11 - m22)) / 2
    qy = math.sqrt(max(0, 1 - m00 + m11 - m22)) / 2
    qz = math.sqrt(max(0, 1 - m00 - m11 + m22)) / 2
    qx = math.copysign(qx, m21 - m12)
    qy = math.copysign(qy, m02 - m20)
    qz = math.copysign(qz, m10 - m01)
    return [qx, qy, qz, qw]


def angle_between(quaternion0, quaternion1):
    # https://math.stackexchange.com/questions/90081/quaternion-distance
    inner_product = sum(a * b for (a, b) in zip(quaternion0, quaternion1))
    return math.acos(2 * inner_product**2 - 1)


# vim: expandtab sw=4 ts=4



import unittest
import datetime
from unittest.mock import MagicMock, patch, call

from subt.radio import Radio
from osgar.bus import Bus
from osgar.lib.serialize import serialize


class RadioTest(unittest.TestCase):

    def test_on_radio(self):
        # virtual world
        data = [b'A0F150L', serialize([13, 'artf', ['TYPE_RESCUE_RANDY', 15982, 104845, 3080]])]
        bus = MagicMock()
        dev = Radio(bus=bus, config={})
        bus.reset_mock()
        dev.on_radio(data)
        bus.publish.assert_called()

        data = [b'A0F150L', serialize([0, 'pose2d', [11896, 7018, -11886]])]
        bus.reset_mock()
        dev.on_radio(data)
        bus.publish.assert_not_called()

    def test_on_breadcrumb(self):
        data = [29.929257253892082, -1.5821685677914703, 1.575092509709292]  # xyz location of newly deployed breadcrumb
        bus = MagicMock()
        dev = Radio(bus=bus, config={})
        bus.reset_mock()
        dev.on_breadcrumb(data)
        bus.publish.assert_called()
        self.assertEqual(bus.method_calls[0],
                         call.publish('radio', serialize([0, 'breadcrumb', [29.929257253892082, -1.5821685677914703, 1.575092509709292]])))

        new_msg = serialize([0, 'breadcrumb', [29.929257253892082, -1.5821685677914703, 1.575092509709292]])
        bus.reset_mock()
        dev.on_radio([b'A1300L', new_msg])
        bus.publish.assert_called()

    def test_msg_pack(self):
        bus = MagicMock()
        dev = Radio(bus=bus, config={})
        bus.reset_mock()
        dev.send_data('dummy_artf', [1, 2, 3])
        bus.publish.assert_called()
        self.assertEqual(bus.method_calls[-1],
                         call.publish('radio', serialize([0, 'dummy_artf', [1, 2, 3]])))
        dev.send_data('dummy_artf', [1, 2, 3])
        self.assertEqual(bus.method_calls[-1],
                         call.publish('radio', serialize([1, 'dummy_artf', [1, 2, 3]])))

    def test_pose3d(self):
        bus = MagicMock()
        dev = Radio(bus=bus, config={})
        bus.reset_mock()
        data = [[-5.999917943872727, 4.999913526023388, 0.12287766016353471], [6.229389211972107e-12, -7.943606405014296e-12, 1.5034993858140517e-12, 1.0]]
        dev.on_pose3d(data)
        bus.publish.assert_not_called()  # not defined sim_time
        dev.on_sim_time_sec(13)
        dev.on_pose3d(data)
        bus.publish.assert_called()
        self.assertEqual(bus.method_calls[-1],
                         call.publish('radio', serialize([1, 'xyz', [13, [-5.999917943872727, 4.999913526023388, 0.12287766016353471]]])))

# vim: expandtab sw=4 ts=4
